##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2022 Stanford University and the Authors
#
# Authors: Peter Eastman
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import math
import tempfile

import numpy as np

from mdtraj import element, load
from mdtraj.core.topology import Topology
from mdtraj.core.trajectory import Trajectory
from mdtraj.testing import eq


def test_convert(get_fn):
    for filename in ["2EQQ.pdb", "4OH9.pdb"]:
        # Load a PDB file.

        traj1 = load(get_fn(filename))
        with tempfile.NamedTemporaryFile(
            suffix=".pdbx",
            mode="w",
            delete=False,
        ) as file:
            # Save it in PDBx/mmCIF format.

            traj1.save(file.name)

            # Load the PDBx/mmCIF file and make the result is identical.

            traj2 = load(file.name)
            assert eq(traj1.n_frames, traj2.n_frames)
            assert eq(traj1.n_atoms, traj2.n_atoms)
            assert eq(traj1.xyz, traj2.xyz)
            assert eq(traj1.unitcell_lengths, traj2.unitcell_lengths)
            assert eq(traj1.unitcell_angles, traj2.unitcell_angles)
            for a1, a2 in zip(traj1.topology.atoms, traj2.topology.atoms):
                assert eq(a1, a2)

            # Try loading just a subset of the frames and atoms.

            traj3 = load(file.name, atom_indices=range(10, 20), stride=2)
            assert eq(traj3.n_frames, math.ceil(traj1.n_frames / 2))
            assert eq(traj3.n_atoms, 10)
            assert eq(traj1.xyz[::2, 10:20], traj3.xyz)
            atoms1 = list(traj1.topology.atoms)[10:20]
            for a1, a2 in zip(atoms1, traj3.topology.atoms):
                assert eq(a1.name, a2.name)


def test_load_pdbx_file(get_fn):
    """Test loading a PDBx/mmCIF file and verifying topology details."""
    for filename in ["8ddg.cif", "8ddg.cif.gz"]:
        traj = load(get_fn(filename))

        # Verify the number of atoms, residues, chains, and bonds
        assert len([a for a in traj.topology.atoms]) == 64, "Unexpected number of atoms"
        assert len([r for r in traj.topology.residues]) == 3, "Unexpected number of residues"
        assert len([c for c in traj.topology.chains]) == 1, "Unexpected number of chains"
        assert len([b for b in traj.topology.bonds]) == 66, "Unexpected number of bonds"

        # Verify atom names and residue names
        atom_names = [a.name for a in traj.topology.atoms]
        assert "CA" in atom_names, "Expected atom 'CA' not found"
        assert "OXT" in atom_names, "Expected atom 'OXT' not found"
        assert "HE2" in atom_names, "Expected atom 'HE2' not found"
        assert "HXT" not in atom_names, "Unexpected atom 'HXT' found"

        residue_names = [r.name for r in traj.topology.residues]
        assert "TYR" in residue_names, "Expected residue 'TYR' not found"
        assert "PHE" in residue_names, "Expected residue 'PHE' not found"
        assert "ALA" not in residue_names, "Unexpected residue 'ALA' found"

        # Verify chain IDs
        chain_ids = [c.chain_id for c in traj.topology.chains]
        assert "A" in chain_ids, "Expected chain ID 'A' not found"

        # Verify unit cell lengths and angles
        assert traj.unitcell_lengths is not None, "Unit cell lengths not found"
        assert traj.unitcell_angles is not None, "Unit cell angles not found"

        # Expected values for unit cell lengths (in nanometers) and angles (in degrees)
        expected_lengths = [
            2.3140,
            0.4840,
            1.9790,
        ]  # Converted from Ångstroms to nanometers
        expected_angles = [90.000, 107.048, 90.000]

        # Use numpy's allclose for comparison with a tolerance
        assert np.allclose(
            traj.unitcell_lengths[0],
            expected_lengths,
            atol=1e-3,
        ), "Unit cell lengths do not match"
        assert np.allclose(
            traj.unitcell_angles[0],
            expected_angles,
            atol=1e-3,
        ), "Unit cell angles do not match"


def create_test_topology():
    """Create a test topology with 3 chains, including a disulfide bond and a non-standard residue."""
    topology = Topology()
    # Chain 1
    chain1 = topology.add_chain("A")
    res1a = topology.add_residue("GLY", chain1, 1)
    n1 = topology.add_atom("N", element.nitrogen, res1a)
    ca1 = topology.add_atom("CA", element.carbon, res1a)
    c1 = topology.add_atom("C", element.carbon, res1a)
    topology.add_bond(n1, ca1)
    topology.add_bond(ca1, c1)
    res2a = topology.add_residue("ALA", chain1, 2)
    n2 = topology.add_atom("N", element.nitrogen, res2a)
    ca2 = topology.add_atom("CA", element.carbon, res2a)
    c2 = topology.add_atom("C", element.carbon, res2a)
    topology.add_bond(c1, n2)
    topology.add_bond(n2, ca2)
    topology.add_bond(ca2, c2)
    res3a = topology.add_residue("CYS", chain1, 3)
    n3 = topology.add_atom("N", element.nitrogen, res3a)
    ca3 = topology.add_atom("CA", element.carbon, res3a)
    cb3 = topology.add_atom("CB", element.carbon, res3a)
    c3 = topology.add_atom("C", element.carbon, res3a)
    sg1 = topology.add_atom("SG", element.sulfur, res3a)
    topology.add_bond(c2, n3)
    topology.add_bond(n3, ca3)
    topology.add_bond(ca3, c3)
    topology.add_bond(ca3, cb3)
    topology.add_bond(cb3, sg1)
    # Chain 2
    chain2 = topology.add_chain("B")
    res1b = topology.add_residue("GLY", chain2, 1)
    n4 = topology.add_atom("N", element.nitrogen, res1b)
    n4h1 = topology.add_atom("H", element.nitrogen, res1b)
    ca4 = topology.add_atom("CA", element.carbon, res1b)
    cah1 = topology.add_atom("HA2", element.hydrogen, res1b)
    cah2 = topology.add_atom("HA3", element.hydrogen, res1b)
    c4 = topology.add_atom("C", element.carbon, res1b)
    o4 = topology.add_atom("O", element.oxygen, res1b)
    topology.add_bond(n4, ca4)
    topology.add_bond(ca4, c4)
    topology.add_bond(ca4, cah1)
    topology.add_bond(ca4, cah2)
    topology.add_bond(c4, o4)
    topology.add_bond(n4, n4h1)
    res2b = topology.add_residue("ALA", chain2, 2)
    n5 = topology.add_atom("N", element.nitrogen, res2b)
    ca5 = topology.add_atom("CA", element.carbon, res2b)
    cb5 = topology.add_atom("CB", element.carbon, res2b)
    c5 = topology.add_atom("C", element.carbon, res2b)
    o5 = topology.add_atom("O", element.oxygen, res2b)
    h1 = topology.add_atom("H", element.hydrogen, res2b)
    h2 = topology.add_atom("HA", element.hydrogen, res2b)
    h3 = topology.add_atom("HB1", element.hydrogen, res2b)
    h4 = topology.add_atom("HB2", element.hydrogen, res2b)
    h5 = topology.add_atom("HB3", element.hydrogen, res2b)
    topology.add_bond(c4, n5)
    topology.add_bond(n5, h1)
    topology.add_bond(ca5, h2)
    topology.add_bond(ca5, cb5)
    topology.add_bond(cb5, h3)
    topology.add_bond(cb5, h4)
    topology.add_bond(cb5, h5)
    topology.add_bond(ca5, c5)
    topology.add_bond(ca5, n5)
    topology.add_bond(c5, o5)
    res3b = topology.add_residue("CYS", chain2, 3)
    n6 = topology.add_atom("N", element.nitrogen, res3b)
    ca6 = topology.add_atom("CA", element.carbon, res3b)
    cb6 = topology.add_atom("CB", element.carbon, res3b)
    c6 = topology.add_atom("C", element.carbon, res3b)
    sg2 = topology.add_atom("SG", element.sulfur, res3b)
    o61 = topology.add_atom("O", element.oxygen, res3b)
    o62 = topology.add_atom("OXT", element.oxygen, res3b)
    topology.add_bond(c5, n6)
    topology.add_bond(n6, ca6)
    topology.add_bond(ca6, c6)
    topology.add_bond(ca6, cb6)
    topology.add_bond(cb6, sg2)
    topology.add_bond(c6, o61)
    topology.add_bond(c6, o62)
    topology.add_bond(sg1, sg2)
    # Chain 3
    chain3 = topology.add_chain("C")
    res1c = topology.add_residue("2HA", chain3, 1)
    cc1 = topology.add_atom("C1", element.carbon, res1c)
    cc2 = topology.add_atom("C2", element.carbon, res1c)
    cc3 = topology.add_atom("C3", element.carbon, res1c)
    ch11 = topology.add_atom("1H1C", element.hydrogen, res1c)
    ch12 = topology.add_atom("2H1C", element.hydrogen, res1c)
    ch31 = topology.add_atom("1H3C", element.hydrogen, res1c)
    ch32 = topology.add_atom("2H3C", element.hydrogen, res1c)
    ho1 = topology.add_atom("H1", element.hydrogen, res1c)
    ho3 = topology.add_atom("H3", element.hydrogen, res1c)
    co1 = topology.add_atom("O1", element.oxygen, res1c)
    co2 = topology.add_atom("O2", element.oxygen, res1c)
    co3 = topology.add_atom("O3", element.oxygen, res1c)
    topology.add_bond(cc1, cc2)
    topology.add_bond(cc2, cc3)
    topology.add_bond(cc1, ch11)
    topology.add_bond(cc1, ch12)
    topology.add_bond(cc2, ch31)
    topology.add_bond(cc2, ch32)
    topology.add_bond(cc1, ho1)
    topology.add_bond(cc3, ho3)
    topology.add_bond(cc1, co1)
    topology.add_bond(cc2, co2)
    topology.add_bond(cc3, co3)
    return topology


def assert_traj_fields_equal(traj1, traj2):
    assert np.allclose(traj1.xyz, traj2.xyz, atol=1e-5), "Positions do not match after write/read"
    assert traj1.n_atoms == traj2.n_atoms, "Number of atoms does not match"
    assert traj1.n_residues == traj2.n_residues, "Number of residues does not match"
    assert traj1.n_chains == traj2.n_chains, "Number of chains does not match"
    assert np.allclose(traj1.unitcell_lengths, traj2.unitcell_lengths, atol=1e-5), "Unit cell lengths do not match"
    assert np.allclose(traj1.unitcell_angles, traj2.unitcell_angles, atol=1e-5), "Unit cell angles do not match"


def bond_set(top):
    return set(tuple(sorted((a1.index, a2.index))) for a1, a2 in top.bonds)


def assert_bonds_equal(top1, top2):
    bonds1 = bond_set(top1)
    bonds2 = bond_set(top2)
    assert bonds1 == bonds2, f"Bonds do not match after write/read: {bonds1} != {bonds2}"


def test_write_and_read_pdbx(tmp_path):
    """Test writing a PDBx file and reading it back,
    checking all fields including bonds, with two chains and a disulfide bond."""
    topology = create_test_topology()
    n_atoms = topology.n_atoms
    xyz = np.zeros((1, n_atoms, 3), dtype=np.float32)
    for i in range(n_atoms):
        xyz[0, i, :] = [i * 0.1, 0.0, 0.0]
    unitcell_lengths = np.array([[2.0, 2.0, 2.0]], dtype=np.float32)
    unitcell_angles = np.array([[90.0, 90.0, 90.0]], dtype=np.float32)
    traj = Trajectory(xyz=xyz, topology=topology, unitcell_lengths=unitcell_lengths, unitcell_angles=unitcell_angles)
    out_path = tmp_path / "test_write.pdbx"
    traj.save(str(out_path))
    loaded = load(str(out_path))
    assert_traj_fields_equal(traj, loaded)
    assert_bonds_equal(traj.topology, loaded.topology)


def test_write_and_read_pdbx_4py5(tmp_path, get_fn):
    """Test writing and reading 4py5.cif, checking all fields including bonds."""
    cif_path = get_fn("4py5.cif")
    traj = load(cif_path)
    out_path = tmp_path / "test_write_4py5.pdbx"
    traj.save(str(out_path))
    loaded = load(str(out_path))
    assert_traj_fields_equal(traj, loaded)
    assert len(bond_set(traj.topology)) == 2928
    assert_bonds_equal(traj.topology, loaded.topology)


def test_write_and_read_pdbx_from_dcd_with_bonds_and_keepids(tmp_path, get_fn):
    """Test writing CIF from DCD+PDB topology with bonds, ensuring bonds are preserved.

    This test addresses the issue where bonds were lost when converting DCD trajectories
    with PDB topologies to CIF format. The problem occurred because:
    1. Chain and residue objects in bonds were different instances than those in topology.chains/residues
    2. Using objects as dictionary keys failed due to Python's object identity hashing

    The fix uses .index properties (unique integers) as dictionary keys instead of objects.
    """
    # Load DCD with PDB topology that has bonds
    dcd_path = get_fn("ala_dipeptide.dcd")
    pdb_path = get_fn("ala_dipeptide.pdb")
    traj = load(dcd_path, top=pdb_path)

    # Verify we have bonds in the original topology
    orig_bonds = len(list(traj.topology.bonds))
    assert orig_bonds > 0, "Original topology has no bonds to test preservation"

    # Save first frame to CIF with keepIds=False (default - auto-generate chain IDs)
    cif_path_no_keepids = tmp_path / "test_dcd_to_cif_no_keepids.cif"
    traj[0].save_cif(str(cif_path_no_keepids), keepIds=False)

    # Load back and verify bonds are preserved
    loaded_no_keepids = load(str(cif_path_no_keepids))
    loaded_bonds_no_keepids = len(list(loaded_no_keepids.topology.bonds))
    assert loaded_bonds_no_keepids == orig_bonds, (
        f"Bonds lost with keepIds=False: {orig_bonds} -> {loaded_bonds_no_keepids}"
    )
    assert_bonds_equal(traj[0].topology, loaded_no_keepids.topology)

    # Save with keepIds=True (preserve original chain IDs from PDB)
    cif_path_keepids = tmp_path / "test_dcd_to_cif_keepids.cif"
    traj[0].save_cif(str(cif_path_keepids), keepIds=True)

    # Load back and verify bonds are preserved
    loaded_keepids = load(str(cif_path_keepids))
    loaded_bonds_keepids = len(list(loaded_keepids.topology.bonds))
    assert loaded_bonds_keepids == orig_bonds, f"Bonds lost with keepIds=True: {orig_bonds} -> {loaded_bonds_keepids}"
    assert_bonds_equal(traj[0].topology, loaded_keepids.topology)


def test_residue_selection_from_cif(get_fn):
    """Tests that residues are correctly parsed as integers in CIF files"""

    # Load the trajectory from the test CIF file using the fixture
    cif_file = get_fn("8ddg.cif")
    traj = load(cif_file)
    top = traj.topology

    # Try selecting by residue number
    residue_atoms = top.select("residue 67")  # ResSeq based index
    resid_atoms = top.select("resid 0")  # 0-based index

    assert residue_atoms.size > 0, "Expected atoms in residue 67 from CIF"
    assert len(residue_atoms) == len(resid_atoms), "Expected same number of atoms in residue 67 and resid 0"
    assert np.array_equal(residue_atoms, resid_atoms), "Expected atoms in residue 67 and resid 0 to be the same"
    assert np.array_equal(
        top.select("residue 68"),
        top.select("resid 1"),
    ), "Expected atoms in residue 68 and resid 1 to be the same"



import itertools

import numpy as np
import pytest
from conftest import flaky_pdb_dl

import mdtraj as md


def call_dssp(get_fn, ref_name, frame=0):
    """Read reference output files created with mkdssp v2.2.1"""
    out = get_fn(f"dssp/{ref_name}_{frame}.dssp")

    KEY_LINE = (
        "  #  RESIDUE AA STRUCTURE BP1 BP2  ACC     N-H-->O    O-->H-N    N-H-->O    "
        "O-->H-N    TCO  KAPPA ALPHA  PHI   PSI    X-CA   Y-CA   Z-CA"
    )
    with open(out) as f:
        # exaust the first entries
        max(itertools.takewhile(lambda line: not line.startswith(KEY_LINE), f))
        return np.array([line[16] for line in f if line[13] != "!"], dtype="<U1")


def assert_(a, b):
    try:
        assert np.all(a == b)
    except AssertionError:
        flag = False  # Flag to see if it's a true error or not (e.g. not PPII)

        if len(a) != len(b):
            print("Not the same length: %d vs %s" % (len(a), len(b)))
            raise

        for i, (aa, bb) in enumerate(zip(a, b)):
            if aa == bb:
                print("%3d: '%s' '%s'" % (i, aa, bb))
            elif aa == "P" and bb == " ":
                # Bypassing cases where mkdssp outputs PPII ("P") and mdtraj doesn't.
                print("%3d: '%s' '%s'" % (i, aa, bb))
            else:
                print("%3d: '%s' '%s' <-" % (i, aa, bb))
                flag = True

        if flag:
            raise


@pytest.mark.parametrize("fn", ["1bpi.pdb", "1vii.pdb", "4ZUO.pdb", "1am7_protein.pdb"])
def test_1(get_fn, tmpdir, fn):
    """This test checks dssp assignments for pdb files in tests/data"""
    t = md.load_pdb(get_fn(fn))
    t = t.atom_slice(t.top.select_atom_indices("minimal"))
    assert_(call_dssp(get_fn, fn), md.compute_dssp(t, simplified=False)[0])


@pytest.mark.parametrize("fn", ["2EQQ.pdb"])
def test_2(get_fn, tmpdir, fn):
    """This test checks dssp assignments on different chains for pdb files in tests/data"""
    t = md.load(get_fn(fn))
    for i in range(len(t)):
        assert_(call_dssp(get_fn, fn, i), md.compute_dssp(t[i], simplified=False)[0])


@flaky_pdb_dl
@pytest.mark.parametrize("pdbid", ["1GAI", "6gsv", "2AAC"])
def test_3(get_fn, tmpdir, pdbid):
    """This test checks dssp assignments on pdb files downloaded from rcsb"""
    # 1COY gives a small error, due to a broken chain.
    t = md.load_pdb("https://www.rcsb.org/pdb/files/%s.pdb" % pdbid)
    t = t.atom_slice(t.top.select_atom_indices("minimal"))
    assert_(call_dssp(get_fn, pdbid), md.compute_dssp(t, simplified=False)[0])


def test_4(get_fn):
    t = md.load_pdb(get_fn("1am7_protein.pdb"))
    a = md.compute_dssp(t, simplified=True)
    b = md.compute_dssp(t, simplified=False)
    assert len(a) == len(b)
    assert len(a[0]) == len(b[0])
    assert list(np.unique(a[0])) == ["C", "E", "H"]


def test_5(get_fn):
    t = md.load(get_fn("4waters.pdb"))
    a = md.compute_dssp(t, simplified=True)
    b = md.compute_dssp(t, simplified=False)
    ref = np.array([["NA", "NA", "NA", "NA"]])

    np.testing.assert_array_equal(a, ref)
    np.testing.assert_array_equal(b, ref)


def test_7(get_fn):
    t = md.load(get_fn("2EQQ.pdb"))
    md.compute_dssp(t, simplified=True)



"""gro.py: Used for loading Gromacs GRO files."""

##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2015 Stanford University and the Authors
#
# Authors: Robert McGibbon, Lee-Ping Wang, Peter Eastman
# Contributors: Jason Swails
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#
# Portions of this code originate from the OpenMM molecular simulation
# toolkit, copyright (c) 2012 Stanford University and the Authors. Those
# portions are distributed under the following terms:
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
##############################################################################

import itertools
import os
import warnings
from re import findall, match, sub

import numpy as np

import mdtraj as md
from mdtraj.core import element as elem
from mdtraj.formats import pdb
from mdtraj.formats.registry import FormatRegistry
from mdtraj.utils import cast_indices, ensure_type, in_units_of


@FormatRegistry.register_loader(".gro")
def load_gro(filename, stride=None, atom_indices=None, frame=None, top=None):
    """Load a GROMACS GRO file.

    Parameters
    ----------
    filename : path-like
        Path to the GRO file on disk.
    stride : int, default=None
        Only read every stride-th model from the file
    atom_indices : array_like, optional
        If not none, then read only a subset of the atoms coordinates from the
        file. These indices are zero-based.
    frame : int, optional
        Use this option to load only a single frame from a trajectory on disk.
        If frame is None, the default, the entire trajectory will be loaded.
        If supplied, ``stride`` will be ignored.
    top : mdtraj.core.Topology, default=None
        if you give a topology as input the topology won't be parsed from the gro file
        it saves time if you have to parse a big number of files
    """

    with GroTrajectoryFile(filename, "r", top=top) as f:
        if frame is not None:
            f.seek(frame)
            n_frames = 1
        else:
            n_frames = None
        return f.read_as_traj(
            n_frames=n_frames,
            stride=stride,
            atom_indices=atom_indices,
        )


@FormatRegistry.register_fileobject(".gro")
class GroTrajectoryFile:
    """Interface for reading and writing to GROMACS GRO files.

    Parameters
    ----------
    filename : path-like
        The filename to open. A path to a file on disk.
    mode : {'r', 'w'}
        The mode in which to open the file, either 'r' for read or 'w' for write.
    force_overwrite : bool
        If opened in write mode, and a file by the name of `filename` already
        exists on disk, should we overwrite it?
    top : mdtraj.core.Topology, default=None
        if you give a topology as input the topology won't be parsed from the gro file
        it saves time if you have to parse a big number of files

    Attributes
    ----------
    n_atoms : int
        The number of atoms in the file
    topology : md.Topology
        The topology. TODO(rmcgibbo) note about chain

    See Also
    --------
    load_gro : High-level wrapper that returns a ``md.Trajectory``
    """

    distance_unit = "nanometers"

    def __init__(self, filename, mode="r", force_overwrite=True, top=None):
        self._open = False
        self._file = None
        self._mode = mode

        self.topology = top

        if mode == "r":
            self._open = True
            self._frame_index = 0
            self._file = open(filename)
            try:
                if self.topology is None:
                    self.n_atoms, self.topology = self._read_topology()
                else:
                    self.n_atoms = self.topology.n_atoms
            finally:
                self._file.seek(0)
        elif mode == "w":
            self._open = True
            if os.path.exists(filename) and not force_overwrite:
                raise OSError('"%s" already exists' % filename)
            self._frame_index = 0
            self._file = open(filename, "w")
        else:
            raise ValueError("invalid mode: %s" % mode)

    def write(
        self,
        coordinates,
        topology,
        time=None,
        unitcell_vectors=None,
        precision=3,
    ):
        """Write one or more frames of a molecular dynamics trajectory to disk
        in the GROMACS GRO format.

        Parameters
        ----------
        coordinates : np.ndarray, dtype=np.float32, shape=(n_frames, n_atoms, 3)
            The cartesian coordinates of each atom, in units of nanometers.
        topology : mdtraj.Topology
            The Topology defining the model to write.
        time : np.ndarray, dtype=float32, shape=(n_frames), optional
            The simulation time corresponding to each frame, in picoseconds.
            If not supplied, the numbers 0..n_frames will be written.
        unitcell_vectors : np.ndarray, dtype=float32, shape=(n_frames, 3, 3), optional
            The periodic box vectors of the simulation in each frame, in nanometers.
        precision : int, optional
            The number of decimal places to print for coordinates. Default is 3
        """
        if not self._open:
            raise ValueError("I/O operation on closed file")
        if not self._mode == "w":
            raise ValueError("file not opened for writing")

        coordinates = ensure_type(
            coordinates,
            dtype=np.float32,
            ndim=3,
            name="coordinates",
            can_be_none=False,
            warn_on_cast=False,
        )
        time = ensure_type(
            time,
            dtype=float,
            ndim=1,
            name="time",
            can_be_none=True,
            shape=(len(coordinates),),
            warn_on_cast=False,
        )
        unitcell_vectors = ensure_type(
            unitcell_vectors,
            dtype=float,
            ndim=3,
            name="unitcell_vectors",
            can_be_none=True,
            shape=(len(coordinates), 3, 3),
            warn_on_cast=False,
        )

        for i in range(coordinates.shape[0]):
            frame_time = None if time is None else time[i]
            frame_box = None if unitcell_vectors is None else unitcell_vectors[i]
            self._write_frame(
                coordinates[i],
                topology,
                frame_time,
                frame_box,
                precision,
            )

    def read_as_traj(self, n_frames=None, stride=None, atom_indices=None):
        """Read a trajectory from a gro file

        Parameters
        ----------
        n_frames : int, optional
            If positive, then read only the next `n_frames` frames. Otherwise read all
            of the frames in the file.
        stride : np.ndarray, optional
            Read only every stride-th frame.
        atom_indices : array_like, optional
            If not none, then read only a subset of the atoms coordinates from the
            file. This may be slightly slower than the standard read because it required
            an extra copy, but will save memory.

        Returns
        -------
        trajectory : Trajectory
            A trajectory object containing the loaded portion of the file.
        """
        from mdtraj.core.trajectory import Trajectory

        topology = self.topology
        if atom_indices is not None:
            topology = topology.subset(atom_indices)

        coordinates, time, unitcell_vectors = self.read(
            stride=stride,
            atom_indices=atom_indices,
        )
        if len(coordinates) == 0:
            return Trajectory(xyz=np.zeros((0, topology.n_atoms, 3)), topology=topology)

        coordinates = in_units_of(
            coordinates,
            self.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )
        unitcell_vectors = in_units_of(
            unitcell_vectors,
            self.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )

        traj = Trajectory(xyz=coordinates, topology=topology, time=time)
        traj.unitcell_vectors = unitcell_vectors
        return traj

    def read(self, n_frames=None, stride=None, atom_indices=None):
        """Read data from a molecular dynamics trajectory in the GROMACS GRO
        format.

        Parameters
        ----------
        n_frames : int, optional
            If n_frames is not None, the next n_frames of data from the file
            will be read. Otherwise, all of the frames in the file will be read.
        stride : int, optional
            If stride is not None, read only every stride-th frame from disk.
        atom_indices : np.ndarray, dtype=int, optional
            The specific indices of the atoms you'd like to retrieve. If not
            supplied, all of the atoms will be retrieved.

        Returns
        -------
        coordinates : np.ndarray, shape=(n_frames, n_atoms, 3)
            The cartesian coordinates of the atoms, in units of nanometers.
        time : np.ndarray, None
            The time corresponding to each frame, in units of picoseconds, or
            None if no time information is present in the trajectory.
        unitcell_vectors : np.ndarray, shape=(n_frames, 3, 3)
            The box vectors in each frame, in units of nanometers
        """
        if not self._open:
            raise ValueError("I/O operation on closed file")
        if not self._mode == "r":
            raise ValueError("file not opened for reading")

        coordinates = []
        unitcell_vectors = []
        time = []
        contains_time = True

        atom_indices = cast_indices(atom_indices)
        atom_slice = slice(None) if atom_indices is None else atom_indices

        if n_frames is None:
            frameiter = itertools.count()
        else:
            frameiter = range(n_frames)

        for i in frameiter:
            try:
                frame_xyz, frame_box, frame_time = self._read_frame()
                contains_time = contains_time and (frame_time is not None)
                coordinates.append(frame_xyz[atom_slice])
                unitcell_vectors.append(frame_box)
                time.append(frame_time)
            except StopIteration:
                break

        coordinates, unitcell_vectors, time = map(
            np.array,
            (coordinates, unitcell_vectors, time),
        )

        if not contains_time:
            time = None
        else:
            time = time[::stride]

        return coordinates[::stride], time, unitcell_vectors[::stride]

    def _read_topology(self):
        if not self._open:
            raise ValueError("I/O operation on closed file")
        if not self._mode == "r":
            raise ValueError("file not opened for reading")
        pdb.PDBTrajectoryFile._loadNameReplacementTables()

        n_atoms = None
        topology = md.Topology()
        chain = topology.add_chain()
        residue = None
        atomReplacements = {}

        # This is needed because sometimes
        # residue names get replaced and this
        # brings to wrong residue parsing
        old_resname = None
        for ln, line in enumerate(self._file):
            if ln == 1:
                n_atoms = int(line.strip())
            elif ln > 1 and ln < n_atoms + 2:
                (thisresnum, thisresname, thisatomname, thisatomnum) = (
                    line[i * 5 : i * 5 + 5].strip() for i in range(4)
                )
                thisresnum, thisatomnum = map(int, (thisresnum, thisatomnum))
                if residue is None or residue.resSeq != thisresnum or old_resname != thisresname:
                    if residue is not None and thisresnum == residue.resSeq:
                        warnings.warn(
                            f"WARNING: two consecutive residues with same number ({thisresname}, {old_resname})",
                        )
                    old_resname = thisresname
                    if thisresname in pdb.PDBTrajectoryFile._residueNameReplacements:
                        thisresname = pdb.PDBTrajectoryFile._residueNameReplacements[thisresname]
                    residue = topology.add_residue(
                        thisresname,
                        chain,
                        resSeq=thisresnum,
                    )
                    if thisresname in pdb.PDBTrajectoryFile._atomNameReplacements:
                        atomReplacements = pdb.PDBTrajectoryFile._atomNameReplacements[thisresname]
                    else:
                        atomReplacements = {}

                thiselem = thisatomname
                if len(thiselem) > 1:
                    thiselem = thiselem[0] + sub("[A-Z0-9]", "", thiselem[1:])
                try:
                    element = elem.get_by_symbol(thiselem)
                except KeyError:
                    element = elem.virtual
                if thisatomname in atomReplacements:
                    thisatomname = atomReplacements[thisatomname]

                topology.add_atom(
                    thisatomname,
                    element=element,
                    residue=residue,
                    serial=thisatomnum,
                )
        topology.create_standard_bonds()
        return n_atoms, topology

    def _read_frame(self):
        if not self._open:
            raise ValueError("I/O operation on closed file")
        if not self._mode == "r":
            raise ValueError("file not opened for reading")

        atomcounter = itertools.count()
        comment = None
        boxvectors = None

        xyz = np.zeros((self.n_atoms, 3), dtype=np.float32)

        got_line = False
        firstDecimalPos = None
        atomindex = -1
        for ln, line in enumerate(self._file):
            got_line = True
            if ln == 0:
                comment = line.strip()
                continue
            elif ln == 1:
                assert self.n_atoms == int(line.strip())
                continue
            if firstDecimalPos is None:
                try:
                    firstDecimalPos = line.index(".", 20)
                    secondDecimalPos = line.index(".", firstDecimalPos + 1)
                except ValueError:
                    firstDecimalPos = secondDecimalPos = None
            crd = _parse_gro_coord(line, firstDecimalPos, secondDecimalPos)
            if crd is not None and atomindex < self.n_atoms - 1:
                atomindex = next(atomcounter)
                xyz[atomindex, :] = (crd[0], crd[1], crd[2])
            elif _is_gro_box(line) and ln == self.n_atoms + 2:
                sline = line.split()
                boxvectors = tuple([float(i) for i in sline])
                # the gro_box line comes at the end of the record
                break
            else:
                raise Exception("Unexpected line in .gro file: " + line)

        if not got_line:
            raise StopIteration()

        time = None
        if "t=" in comment:
            # title string (free format string, optional time in ps after 't=')
            time = float(findall(r"t= *(\d+\.\d+)", comment)[-1])

        # box vectors (free format, space separated reals), values: v1(x) v2(y)
        # v3(z) v1(y) v1(z) v2(x) v2(z) v3(x) v3(y), the last 6 values may be
        # omitted (they will be set to zero).
        box = [boxvectors[i] if i < len(boxvectors) else 0 for i in range(9)]
        unitcell_vectors = np.array(
            [
                [box[0], box[3], box[4]],
                [box[5], box[1], box[6]],
                [box[7], box[8], box[2]],
            ],
        )

        return xyz, unitcell_vectors, time

    def _write_frame(self, coordinates, topology, time, box, precision):
        comment = "Generated with MDTraj"
        if time is not None:
            comment += ", t= %s" % time

        varwidth = precision + 5
        fmt = "%%5d%%-5s%%5s%%5d%%%d.%df%%%d.%df%%%d.%df" % (
            varwidth,
            precision,
            varwidth,
            precision,
            varwidth,
            precision,
        )
        assert topology.n_atoms == coordinates.shape[0]
        lines = [comment, " %d" % topology.n_atoms]
        if box is None:
            box = np.zeros((3, 3))

        for i in range(topology.n_atoms):
            atom = topology.atom(i)
            residue = atom.residue
            serial = atom.serial
            if serial is None:
                serial = atom.index
            if serial >= 100000:
                serial %= 100000
            lines.append(
                fmt
                % (
                    residue.resSeq,
                    residue.name,
                    atom.name,
                    serial,
                    coordinates[i, 0],
                    coordinates[i, 1],
                    coordinates[i, 2],
                ),
            )

        lines.append(
            f"{box[0, 0]:10.5f}{box[1, 1]:10.5f}{box[2, 2]:10.5f}"
            f"{box[0, 1]:10.5f}{box[0, 2]:10.5f}{box[1, 0]:10.5f}"
            f"{box[1, 2]:10.5f}{box[2, 0]:10.5f}{box[2, 1]:10.5f}",
        )

        self._file.write("\n".join(lines))
        self._file.write("\n")

    def seek(self, offset, whence=0):
        """Move to a new file position

        Parameters
        ----------
        offset : int
            A number of frames.
        whence : {0, 1, 2}
            0: offset from start of file, offset should be >=0.
            1: move relative to the current position, positive or negative
            2: move relative to the end of file, offset should be <= 0.
            Seeking beyond the end of a file is not supported
        """
        raise NotImplementedError()

    def tell(self):
        """Current file position

        Returns
        -------
        offset : int
            The current frame in the file.
        """
        return self._frame_index

    def close(self):
        "Close the file"
        if self._open:
            self._file.close()
            self._open = False

    def __enter__(self):
        "Support the context manager protocol"
        return self

    def __exit__(self, *exc_info):
        "Support the context manager protocol"
        self.close()


def _isint(word):
    """ONLY matches integers! If you have a decimal point? None shall pass!

    @param[in] word String (for instance, '123', '153.0', '2.', '-354')
    @return answer Boolean which specifies whether the string is an integer (only +/- sign followed by digits)

    """
    return match("^[-+]?[0-9]+$", word)


def _isfloat(word):
    """Matches ANY number; it can be a decimal, scientific notation, what have you
    CAUTION - this will also match an integer.

    @param[in] word String (for instance, '123', '153.0', '2.', '-354')
    @return answer Boolean which specifies whether the string is any number

    """
    return match(r"^[-+]?[0-9]*\.?[0-9]*([eEdD][-+]?[0-9]+)?$", word)


def _parse_gro_coord(line, firstDecimal, secondDecimal):
    """Determines whether a line contains GROMACS data or not

    @param[in] line The line to be tested

    """
    if firstDecimal is None or secondDecimal is None:
        return None
    digits = secondDecimal - firstDecimal
    try:
        return tuple(float(line[20 + i * digits : 20 + (i + 1) * digits]) for i in range(3))
    except ValueError:
        return None


def _is_gro_box(line):
    """Determines whether a line contains a GROMACS box vector or not

    @param[in] line The line to be tested

    """
    sline = line.split()
    if len(sline) == 9 and all([_isfloat(i) for i in sline]):
        return 1
    elif len(sline) == 3 and all([_isfloat(i) for i in sline]):
        return 1
    else:
        return 0



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Matthew Harrigan
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import itertools
import os

import numpy as np
import pytest

import mdtraj as md
from mdtraj import element
from mdtraj.utils.unitcell import check_valid_unitcell_angles

flaky_pdb_dl = pytest.mark.flaky(rerun=3, reason="github-node flaky pdb dl")


@pytest.fixture(scope="session")
def get_fn():
    test_dir = os.path.dirname(os.path.abspath(__file__))

    def _get_fn(fn):
        return f"{test_dir}/data/{fn}"

    return _get_fn


@pytest.fixture(scope="function")
def gen_random_ptraj(request):
    """Fixture for preparing a test trajectories with random coordinates
    and unitcell
    """
    request.cls.N_FRAMES = N_FRAMES = 20
    request.cls.N_ATOMS = N_ATOMS = 20

    request.cls.rng = rng = np.random.default_rng()

    request.cls.xyz = xyz = np.asarray(rng.standard_normal((N_FRAMES, N_ATOMS, 3), dtype=np.float32))
    request.cls.pairs = np.array(list(itertools.combinations(range(N_ATOMS), 2)), dtype=np.int32)
    request.cls.times = np.array([[i, 0] for i in range(N_FRAMES)[::2]], dtype=np.int32)

    request.cls.ptraj = md.Trajectory(xyz=xyz, topology=None)
    request.cls.ptraj.unitcell_vectors = np.ascontiguousarray(
        rng.standard_normal((N_FRAMES, 3, 3)) + 2 * np.eye(3, 3),
        dtype=np.float32,
    )

    while True:
        # RNG can give us unitcells with zero or imaginary volume.
        # This `while` loop guarantees that we don't get that.
        request.cls.ptraj.unitcell_vectors = np.ascontiguousarray(
            rng.standard_normal((N_FRAMES, 3, 3)) + 2 * np.eye(3, 3),
            dtype=np.float32,
        )

        if np.all([check_valid_unitcell_angles(*row) for row in request.cls.ptraj.unitcell_angles]):
            break


@pytest.fixture(scope="function")
def h5traj(tmp_path):
    xyz = np.around(np.random.randn(10, 5, 3).astype(np.float32), 2)
    topology = md.Topology()
    chain = topology.add_chain()
    residue = topology.add_residue("ALA", chain)
    topology.add_atom("CA", element.carbon, residue)
    topology.add_atom("HG1", element.hydrogen, residue)
    topology.add_atom("SG", element.sulfur, residue)
    topology.add_atom("OD1", element.oxygen, residue)
    topology.add_atom("NE", element.nitrogen, residue)

    time = np.arange(10) ** 2
    unitcell_lengths = np.array([[1.1, 1.2, 1.3]] * 10)
    unitcell_angles = np.array([[90, 90, 95]] * 10)

    traj = md.Trajectory(
        xyz,
        topology=topology,
        time=time,
        unitcell_lengths=unitcell_lengths,
        unitcell_angles=unitcell_angles,
    )

    fn = f"{tmp_path}/ref.h5"
    traj.save(fn)
    return traj, fn, str(tmp_path)


@pytest.fixture(scope="function")
def h5traj_full_metadata(tmp_path):
    # h5 trajectory with full bond metadata and formal charge
    # for use testing roundtrip of topologies
    # added because fixture above  (h5traj) is used
    # in many tests where full data is not roundtripped.
    # PR2101

    from mdtraj.core.topology import Double, Single

    xyz = np.around(np.random.randn(10, 5, 3).astype(np.float32), 2)
    topology = md.Topology()
    chain = topology.add_chain()
    residue = topology.add_residue("ALA", chain)
    ca = topology.add_atom("CA", element.carbon, residue, formal_charge=0)
    hg1 = topology.add_atom("HG1", element.hydrogen, residue, formal_charge=1)
    sg = topology.add_atom("SG", element.sulfur, residue, formal_charge=-1)
    od1 = topology.add_atom("OD1", element.oxygen, residue, formal_charge=-1)
    ne = topology.add_atom("NE", element.nitrogen, residue, formal_charge=1)
    topology.add_bond(ca, hg1, order=1, type=Single)
    topology.add_bond(ca, sg, order=2, type=Double)
    topology.add_bond(od1, ne, order=1, type=Single)

    time = np.arange(10) ** 2
    unitcell_lengths = np.array([[1.1, 1.2, 1.3]] * 10)
    unitcell_angles = np.array([[90, 90, 95]] * 10)

    traj = md.Trajectory(
        xyz,
        topology=topology,
        time=time,
        unitcell_lengths=unitcell_lengths,
        unitcell_angles=unitcell_angles,
    )

    fn = f"{tmp_path}/ref.h5"
    traj.save(fn)
    return traj, fn, str(tmp_path)



import json
import os
import shutil
import subprocess
import sys
import tempfile
import warnings
from distutils.ccompiler import new_compiler
from distutils.dep_util import newer_group
from distutils.errors import DistutilsExecError, DistutilsSetupError
from distutils.sysconfig import customize_compiler, get_config_vars

from setuptools import Extension
from setuptools.command.build_ext import build_ext as _build_ext

################################################################################
# Detection of compiler capabilities
################################################################################


class CompilerDetection:
    # Necessary for OSX. See https://github.com/mdtraj/mdtraj/issues/576
    # The problem is that distutils.sysconfig.customize_compiler()
    # is necessary to properly invoke the correct compiler for this class
    # (otherwise the CC env variable isn't respected). Unfortunately,
    # distutils.sysconfig.customize_compiler() DIES on OSX unless some
    # appropriate initialization routines have been called. This line
    # has a side effect of calling those initialzation routes, and is therefor
    # necessary for OSX, even though we don't use the result.
    _DONT_REMOVE_ME = get_config_vars()

    def __init__(self, disable_openmp):
        self.disable_openmp = disable_openmp
        self._is_initialized = False

    def initialize(self):
        if self._is_initialized:
            return

        cc = new_compiler()
        customize_compiler(cc)

        self.msvc = cc.compiler_type == "msvc"
        self._print_compiler_version(cc)

        if self.disable_openmp:
            self.openmp_enabled = False
        else:
            self.openmp_enabled, openmp_needs_gomp = self._detect_openmp()
        self.sse2_enabled = self._detect_sse2() if not self.msvc else True
        self.sse3_enabled = self._detect_sse3() if not self.msvc else True
        self.sse41_enabled = self._detect_sse41() if not self.msvc else True
        self.neon_enabled = self._detect_neon() if not self.msvc else False

        if self.msvc:
            self.compiler_args_sse2 = ["/arch:SSE2"]
        elif self.sse2_enabled:
            self.compiler_args_sse2 = ["-msse2"]
        else:
            self.compiler_args_sse2 = []
        self.compiler_args_sse3 = ["-mssse3"] if (self.sse3_enabled and not self.msvc) else []
        self.compiler_args_neon = []
        self.compiler_args_warn = (
            ["-Wno-unused-function", "-Wno-unreachable-code", "-Wno-sign-compare"] if not self.msvc else []
        )

        if self.neon_enabled:
            self.compiler_args_sse2 = []
            self.compiler_args_sse3 = []

        self.compiler_args_sse41, self.define_macros_sse41 = [], []
        if self.sse41_enabled:
            self.define_macros_sse41 = [("__SSE4__", 1), ("__SSE4_1__", 1)]
            if not self.msvc:
                self.compiler_args_sse41 = ["-msse4"]

        if os.environ.get("MDTRAJ_BUILD_DISABLE_INTRINSICS") == "1":
            self.disable_intrinsics = True
            print("Env var MDTRAJ_BUILD_DISABLE_INTRINSICS set")
        else:
            self.disable_intrinsics = False

        if self.openmp_enabled:
            self.compiler_libraries_openmp = []

            if self.msvc:
                self.compiler_args_openmp = ["/openmp"]
            else:
                self.compiler_args_openmp = ["-fopenmp"]
                if openmp_needs_gomp:
                    self.compiler_libraries_openmp = ["gomp"]
        else:
            self.compiler_libraries_openmp = []
            self.compiler_args_openmp = []

        if self.msvc:
            self.compiler_args_opt = ["/O2"]
        else:
            self.compiler_args_opt = ["-O3", "-funroll-loops", "--std=c++11"]
        print()
        self._is_initialized = True

    def _print_compiler_version(self, cc):
        print("C compiler:")
        try:
            if self.msvc:
                if not cc.initialized:
                    cc.initialize()
                cc.spawn([cc.cc])
            else:
                cc.spawn([cc.compiler[0]] + ["-v"])
        except DistutilsExecError:
            pass

    def hasfunction(self, funcname, include=None, libraries=None, extra_postargs=None):
        # running in a separate subshell lets us prevent unwanted stdout/stderr
        part1 = f"""
from __future__ import print_function
import os
import json
from distutils.ccompiler import new_compiler
from distutils.sysconfig import customize_compiler, get_config_vars

FUNCNAME = json.loads('{json.dumps(funcname)}')
INCLUDE = json.loads('{json.dumps(include)}')
LIBRARIES = json.loads('{json.dumps(libraries or [])}')
EXTRA_POSTARGS = json.loads('{json.dumps(extra_postargs)}')
        """

        part2 = """
get_config_vars()  # DON'T REMOVE ME
cc = new_compiler()
customize_compiler(cc)
for library in LIBRARIES:
    cc.add_library(library)

status = 0
try:
    with open('func.c', 'w') as f:
        if INCLUDE is not None:
            f.write('#include %s\\n' % INCLUDE)
        f.write('int main(void) {\\n')
        f.write('    %s;\\n' % FUNCNAME)
        f.write('}\\n')
    objects = cc.compile(['func.c'], output_dir='.',
                         extra_postargs=EXTRA_POSTARGS)
    cc.link_executable(objects, 'a.out')
except Exception as e:
    status = 1
exit(status)
        """
        tmpdir = tempfile.mkdtemp(prefix="hasfunction-")
        try:
            curdir = os.path.abspath(os.curdir)
            os.chdir(tmpdir)
            with open("script.py", "w") as f:
                f.write(part1 + part2)
            proc = subprocess.Popen(
                [sys.executable, "script.py"],
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            proc.communicate()
            status = proc.wait()
        finally:
            os.chdir(curdir)
            shutil.rmtree(tmpdir)

        return status == 0

    def _print_support_start(self, feature):
        print(f"Attempting to autodetect {feature:6} support...", end=" ")

    def _print_support_end(self, feature, status):
        if status is True:
            print(f"Compiler supports {feature}")
        else:
            print(f"Did not detect {feature} support")

    def _detect_openmp(self):
        self._print_support_start("OpenMP")
        extra_postargs = ["/openmp"] if self.msvc else ["-fopenmp"]
        args = dict(extra_postargs=extra_postargs, include="<omp.h>")
        hasopenmp = self.hasfunction("omp_get_num_threads()", **args)
        needs_gomp = False
        if not hasopenmp:
            hasopenmp = self.hasfunction(
                "omp_get_num_threads()",
                libraries=["gomp"],
                **args,
            )
            needs_gomp = hasopenmp
        self._print_support_end("OpenMP", hasopenmp)
        return hasopenmp, needs_gomp

    def _detect_sse2(self):
        "Does this compiler support SSE2 intrinsics?"
        self._print_support_start("SSE2")
        result = self.hasfunction(
            "__m128d v; _mm_add_pd(v,v)",
            include="<immintrin.h>",  # Possibly emmintrin.h
            extra_postargs=["-msse2"],
        )
        self._print_support_end("SSE2", result)
        return result

    def _detect_sse3(self):
        "Does this compiler support SSE3 intrinsics?"
        self._print_support_start("SSE3")
        result = self.hasfunction(
            "__m128 v; _mm_hadd_ps(v,v)",
            include="<pmmintrin.h>",
            extra_postargs=["-msse3"],
        )
        self._print_support_end("SSE3", result)
        return result

    def _detect_sse41(self):
        "Does this compiler support SSE4.1 intrinsics?"
        self._print_support_start("SSE4.1")
        result = self.hasfunction(
            "__m128 v; _mm_round_ps(v,0x00)",
            include="<smmintrin.h>",
            extra_postargs=["-msse4"],
        )
        self._print_support_end("SSE4.1", result)
        return result

    def _detect_neon(self):
        """Does this compiler support NEON intrinsics (ARM64)"""
        self._print_support_start("NEON")
        result = self.hasfunction(
            "int16x4_t acc = vdup_n_s16(0);",
            include="<arm_neon.h>",
        )
        self._print_support_end("NEON", result)
        return result


################################################################################
# Writing version control information to the module
################################################################################


def git_version():
    # Return the git revision as a string
    # copied from numpy setup.py
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            env=env,
        ).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(["git", "rev-parse", "HEAD"])
        GIT_REVISION = out.strip().decode("ascii")
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def copy_same_file_pass(source, destination):
    try:
        # Try to copy
        shutil.copy(source, destination)
    except shutil.SameFileError:
        # Ignore the error
        pass


class StaticLibrary(Extension):
    def __init__(self, *args, **kwargs):
        self.export_include = kwargs.pop("export_include", [])
        Extension.__init__(self, *args, **kwargs)


class build_ext(_build_ext):
    def build_extension(self, ext):
        if isinstance(ext, StaticLibrary):
            self.build_static_extension(ext)
        else:
            _build_ext.build_extension(self, ext)

    def copy_extensions_to_source(self):
        _extensions = self.extensions
        self.extensions = [e for e in _extensions if not isinstance(e, StaticLibrary)]
        _build_ext.copy_extensions_to_source(self)
        self.extensions = _extensions

    def build_static_extension(self, ext):
        from distutils import log

        sources = ext.sources
        if sources is None or not isinstance(sources, (list, tuple)):
            raise DistutilsSetupError(
                f"in 'ext_modules' option (extension '{ext.name}'), "
                "'sources' must be present and must be "
                "a list of source filenames",
            )
        sources = list(sources)

        ext_path = self.get_ext_fullpath(ext.name)
        depends = sources + ext.depends
        if not (self.force or newer_group(depends, ext_path, "newer")):
            log.debug("skipping '%s' extension (up-to-date)", ext.name)
            return
        else:
            log.info("building '%s' extension", ext.name)

        extra_args = ext.extra_compile_args or []
        macros = ext.define_macros[:]
        for undef in ext.undef_macros:
            macros.append((undef,))
        objects = self.compiler.compile(
            sources,
            output_dir=self.build_temp,
            macros=macros,
            include_dirs=ext.include_dirs,
            debug=self.debug,
            extra_postargs=extra_args,
            depends=ext.depends,
        )
        self._built_objects = objects[:]
        if ext.extra_objects:
            objects.extend(ext.extra_objects)
        extra_args = ext.extra_link_args or []

        language = ext.language or self.compiler.detect_language(sources)

        libname = os.path.basename(ext_path).split(os.extsep)[0]
        output_dir = os.path.dirname(ext_path)

        if self.compiler.static_lib_format.startswith("lib") and libname.startswith(
            "lib",
        ):
            libname = libname[3:]

        # 1. copy to build directory
        # 1. copy to src tree for develop mode
        # 1. catch errors because we're running in editable mode
        import re

        src_tree_output_dir = re.match("build.*(mdtraj.*)", output_dir)

        try:
            # This might fail if no match in editable mode.
            src_tree_output_dir = src_tree_output_dir.group(1)
            if not os.path.exists(src_tree_output_dir):
                os.makedirs(src_tree_output_dir)
        except AttributeError:
            src_tree_output_dir = output_dir

        if not os.path.exists(output_dir):
            # necessary for windows
            os.makedirs(output_dir)

        assert os.path.isdir(src_tree_output_dir)

        self.compiler.create_static_lib(
            objects,
            output_libname=libname,
            output_dir=output_dir,
            target_lang=language,
        )

        lib_path = self.compiler.library_filename(libname, output_dir=output_dir)

        copy_same_file_pass(lib_path, src_tree_output_dir)

        for item in ext.export_include:
            copy_same_file_pass(item, src_tree_output_dir)
            copy_same_file_pass(item, output_dir)

    def get_ext_filename(self, ext_name):
        filename = _build_ext.get_ext_filename(self, ext_name)

        try:
            exts = [e for e in self.extensions if ext_name in {e.name, e.name.split(".")[-1]}]
            ext = exts[0]
            if isinstance(ext, StaticLibrary):
                if new_compiler().compiler_type == "msvc":
                    return filename.split(".")[0] + ".lib"
                else:
                    return filename.split(".")[0] + ".a"
        except Exception as e:  # noqa
            pass
        return filename


def parse_setuppy_commands():
    """Check the commands and respond appropriately.
    Return a boolean value for whether or not to run the build or not (avoid
    parsing Cython and template files if False).

    Adopted from scipy setup
    """
    args = sys.argv[1:]

    if not args:
        # User forgot to give an argument probably, let setuptools handle that.
        return True

    info_commands = [
        "--help-commands",
        "--name",
        "--version",
        "-V",
        "--fullname",
        "--author",
        "--author-email",
        "--maintainer",
        "--maintainer-email",
        "--contact",
        "--contact-email",
        "--url",
        "--license",
        "--description",
        "--long-description",
        "--platforms",
        "--classifiers",
        "--keywords",
        "--provides",
        "--requires",
        "--obsoletes",
    ]

    for command in info_commands:
        if command in args:
            return False

    # Note that 'alias', 'saveopts' and 'setopt' commands also seem to work
    # fine as they are, but are usually used together with one of the commands
    # below and not standalone.  Hence they're not added to good_commands.
    good_commands = (
        "develop",
        "sdist",
        "build",
        "build_ext",
        "build_py",
        "build_clib",
        "build_scripts",
        "bdist_wheel",
        "bdist_rpm",
        "bdist_wininst",
        "bdist_msi",
        "bdist_mpkg",
        "build_sphinx",
    )

    for command in good_commands:
        if command in args:
            return True

    # The following commands are supported, but we need to show more
    # useful messages to the user
    if "install" in args:
        return True

    if "--help" in args or "-h" in sys.argv[1]:
        return False

    # Commands that do more than print info, but also don't need Cython and
    # template parsing.
    other_commands = ["egg_info", "install_egg_info", "rotate"]
    for command in other_commands:
        if command in args:
            return False

    # If we got here, we didn't detect what setup.py command was given
    warnings.warn(
        "Unrecognized setuptools command ('{}'), proceeding with "
        "generating Cython sources and expanding templates".format(
            " ".join(sys.argv[1:]),
        ),
    )
    return True



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Kyle A. Beauchamp
# Contributors: Robert McGibbon, John D. Chodera
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#
# Portions of this code originate from the OpenMM molecular simulation
# toolkit, copyright (c) 2012 Stanford University and Peter Eastman. Those
# portions are distributed under the following terms:
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
##############################################################################
"""Load an md.Topology from tripos mol2 files."""

import itertools
import re
from io import StringIO

import numpy as np

from mdtraj.core import element as elem
from mdtraj.formats.registry import FormatRegistry
from mdtraj.utils import import_

__all__ = ["load_mol2", "mol2_to_dataframes"]


@FormatRegistry.register_loader(".mol2")
def load_mol2(filename):
    """Load a TRIPOS mol2 file from disk.

    Parameters
    ----------
    filename : path-like
        Path to the prmtop file on disk.

    Returns
    -------
    traj : md.Trajectory
        The resulting topology, as an md.Topology object.

    Notes
    -----
    This function should work on GAFF and sybyl style MOL2 files, but has
    been primarily tested on GAFF mol2 files.
    This function does NOT accept multi-structure MOL2 files!!!
    The elements are guessed using GAFF atom types or via the atype string.

    Examples
    --------
    >>> traj = md.load_mol2('mysystem.mol2')
    """
    from mdtraj.core.topology import Amide, Aromatic, Double, Single, Topology, Triple
    from mdtraj.core.trajectory import Trajectory

    atoms, bonds = mol2_to_dataframes(filename)

    atoms_mdtraj = atoms[["name", "resName"]].copy()
    atoms_mdtraj["serial"] = atoms.index

    # Figure out 1 letter element names

    # IF this is a GAFF mol2, this line should work without issues
    atoms_mdtraj["element"] = atoms.atype.map(gaff_elements)
    # If this is a sybyl mol2, there should be NAN (null) values
    if atoms_mdtraj.element.isnull().any():
        # If this is a sybyl mol2, I think this works generally.
        # Argument x is being passed as a list with only one element.
        def to_element(x):
            if isinstance(x, (list, tuple)):
                assert len(x) == 1
                x = x[0]

            if "." in x:  # orbital-hybridizations in SYBL
                return x.split(".")[0]
            try:
                # check if we can convert the whole str to an Element,
                # if not, we only pass the first letter.
                from mdtraj.core.element import Element

                Element.getBySymbol(x)
            except KeyError:
                return x[0]
            return x

        atoms_mdtraj["element"] = atoms.atype.apply(to_element)

    # Check if elements inferred from atoms.atype are valid
    # If not, try to infer elements from atoms.name
    try:
        atoms_mdtraj["element"].apply(elem.get_by_symbol)
    except KeyError:
        try:
            atoms_mdtraj["element"] = atoms.name.apply(to_element)
            atoms_mdtraj["element"].apply(elem.get_by_symbol)
        except KeyError:
            raise KeyError("Invalid element passed to atoms DataFrame")

    atoms_mdtraj["resSeq"] = atoms["code"]
    atoms_mdtraj["chainID"] = np.ones(len(atoms), "int")

    bond_type_map = {
        "1": Single,
        "2": Double,
        "3": Triple,
        "am": Amide,
        "ar": Aromatic,
    }
    if bonds is not None:
        bonds_mdtraj = bonds[["id0", "id1"]].values
        offset = bonds_mdtraj.min()  # Should this just be 1???
        bonds_mdtraj -= offset
        # Create the bond augment information
        n_bonds = bonds_mdtraj.shape[0]
        bond_augment = np.zeros([n_bonds, 2], dtype=float)
        # Add bond type information
        bond_augment[:, 0] = [float(bond_type_map[str(bond_value)]) for bond_value in bonds["bond_type"].values]
        # Add Bond "order" information, this is not known from Mol2 files
        bond_augment[:, 1] = [0.0 for _ in range(n_bonds)]
        # Augment array, dtype is cast to minimal representation of float
        bonds_mdtraj = np.append(bonds_mdtraj, bond_augment, axis=-1)
    else:
        bonds_mdtraj = None

    top = Topology.from_dataframe(atoms_mdtraj, bonds_mdtraj)

    xyzlist = np.array([atoms[["x", "y", "z"]].values])
    xyzlist /= 10.0  # Convert from angstrom to nanometer

    traj = Trajectory(xyzlist, top)

    return traj


def mol2_to_dataframes(filename):
    """Convert a GAFF (or sybyl) mol2 file to a pair of pandas dataframes.

    Parameters
    ----------
    filename : path-like
        Name of mol2 filename

    Returns
    -------
    atoms_frame : pd.DataFrame
        DataFrame containing atom information
    bonds_frame : pd.DataFrame
        DataFrame containing bond information

    Notes
    -----
    These dataframes may contain force field information as well as the
    information necessary for constructing the coordinates and molecular
    topology.  This function has been tested for GAFF and sybyl-style
    mol2 files but has been primarily tested on GAFF mol2 files.
    This function does NOT accept multi-structure MOL2 files!!!

    See Also
    --------
    If you just need the coordinates and bonds, use load_mol2(filename)
    to get a Trajectory object.
    """
    pd = import_("pandas")
    with open(filename) as f:
        data = {key: list(grp) for key, grp in itertools.groupby(f, _parse_mol2_sections)}

    # Mol2 can have "status bits" at the end of the bond lines. We don't care
    # about these, but they interfere with using pd_read_table because it looks
    # like one line has too many columns. So we just regex out the offending
    # text.
    status_bit_regex = r"BACKBONE|DICT|INTERRES|\|"
    data["@<TRIPOS>BOND\n"] = [re.sub(status_bit_regex, lambda _: "", s) for s in data["@<TRIPOS>BOND\n"]]

    if len(data["@<TRIPOS>BOND\n"]) > 1:
        csv = StringIO()
        csv.writelines(data["@<TRIPOS>BOND\n"][1:])
        csv.seek(0)
        bonds_frame = pd.read_table(
            csv,
            names=["bond_id", "id0", "id1", "bond_type"],
            index_col=0,
            header=None,
            sep=r"\s+",
            engine="python",
        )
    else:
        bonds_frame = None

    csv = StringIO()
    csv.writelines(data["@<TRIPOS>ATOM\n"][1:])
    csv.seek(0)
    atoms_frame = pd.read_csv(csv, sep=r"\s+", engine="python", header=None)
    ncols = atoms_frame.shape[1]
    names = [
        "serial",
        "name",
        "x",
        "y",
        "z",
        "atype",
        "code",
        "resName",
        "charge",
        "status",
    ]
    atoms_frame.columns = names[:ncols]

    return atoms_frame, bonds_frame


def _parse_mol2_sections(x):
    """Helper function for parsing a section in a MOL2 file."""
    if x.startswith("@<TRIPOS>"):
        _parse_mol2_sections.key = x
    return _parse_mol2_sections.key


gaff_elements = {
    "br": "Br",
    "c": "C",
    "c1": "C",
    "c2": "C",
    "c3": "C",
    "ca": "C",
    "cc": "C",
    "cd": "C",
    "ce": "C",
    "cf": "C",
    "cg": "C",
    "ch": "C",
    "cl": "Cl",
    "cp": "C",
    "cq": "C",
    "cu": "C",
    "cv": "C",
    "cx": "C",
    "cy": "C",
    "cz": "C",
    "f": "F",
    "h1": "H",
    "h2": "H",
    "h3": "H",
    "h4": "H",
    "h5": "H",
    "ha": "H",
    "hc": "H",
    "hn": "H",
    "ho": "H",
    "hp": "H",
    "hs": "H",
    "hw": "H",
    "hx": "H",
    "i": "I",
    "n": "N",
    "n1": "N",
    "n2": "N",
    "n3": "N",
    "n4": "N",
    "na": "N",
    "nb": "N",
    "nc": "N",
    "nd": "N",
    "ne": "N",
    "nf": "N",
    "nh": "N",
    "no": "N",
    "o": "O",
    "oh": "O",
    "os": "O",
    "ow": "O",
    "p2": "P",
    "p3": "P",
    "p4": "P",
    "p5": "P",
    "pb": "P",
    "px": "P",
    "py": "P",
    "s": "S",
    "s2": "S",
    "s4": "S",
    "s6": "S",
    "sh": "S",
    "ss": "S",
    "sx": "S",
    "sy": "S",
}



"""
Execute each notebook as a test, reporting an error if any cell throws an exception.
Adapted from https://gist.github.com/minrk/2620876.
"""

import os
import shutil
import socket
import sys
from queue import Empty

import pytest

try:
    import nbformat
    from jupyter_client import KernelManager
except Exception:
    pytest.skip("Skipping no nbformat/jupyter", allow_module_level=True)

FLAKEY_LIST = ["centroids.ipynb", "native-contact.ipynb", "hbonds.ipynb"]
SPARTA_PLUS = ["sparta+", "SPARTA+", "SPARTA+.linux"]
TIMEOUT = 60  # seconds

test_dir = os.path.dirname(os.path.abspath(__file__))
examples = [
    pytest.param(fn, marks=pytest.mark.flaky) if fn in FLAKEY_LIST else fn
    for fn in os.listdir(test_dir)
    if fn.endswith(".ipynb")
]


def is_network_connected():
    try:
        # connect to the host -- tells us if the host is actually
        # reachable
        socket.create_connection(("1.1.1.1", 53))
        return True
    except OSError:
        pass
    return False


def find_executable(names):
    for possible in names:
        result = shutil.which(possible)
        if result is not None:
            return result
    return None


@pytest.fixture(params=examples)
def example_fn(request):
    if "openmm" in request.param:
        try:
            from openmm import app  # noqa: F401
        except ImportError:
            pytest.skip(
                f"Openmm required for example notebook `{request.param}`",
            )

    if "nmr" in request.param:
        if find_executable(SPARTA_PLUS) is None:
            pytest.skip(
                f"Sparta+ not found for example notebook `{request.param}`",
            )

    if not is_network_connected():
        if any(x in request.param for x in ("native-contact", "hbonds")):
            pytest.skip("Network access required")

    cwd = os.path.abspath(".")
    os.chdir(test_dir)
    yield request.param
    os.chdir(cwd)


def test_example_notebook(example_fn):
    with open(example_fn) as f:
        nb = nbformat.reads(f.read(), nbformat.NO_CONVERT)
    run_notebook(nb)


def run_notebook(nb):
    km = KernelManager()
    km.start_kernel(stderr=open(os.devnull, "w"))
    kc = km.client()
    kc.start_channels()
    # simple ping:
    kc.execute("pass")
    kc.get_shell_msg()

    failures = 0
    for cell in nb.cells:
        if cell.cell_type != "code":
            continue
        kc.execute(cell.source)
        try:
            # wait for finish, w/ timeout
            reply = kc.get_shell_msg(timeout=TIMEOUT)["content"]
        except Empty:
            raise Exception(
                f'Timeout ({TIMEOUT:.1f}) when executing the following {cell.cell_type} cell: "{cell.source.strip()}"',
            )
        if reply["status"] == "error":
            failures += 1
            print("\nFAILURE:", file=sys.stderr)
            print("\n".join(reply["traceback"]), file=sys.stderr)
            print(file=sys.stderr)

    kc.stop_channels()
    km.shutdown_kernel()
    del km
    if failures > 0:
        raise Exception()



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2015 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors: Kyle A Beauchamp, Jason Swails
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import numpy as np

from mdtraj.utils import ensure_type
from mdtraj.utils.unitcell import box_vectors_to_lengths_and_angles

from . import _geometry

__all__ = [
    "compute_distances_core",
    "compute_distances",
    "compute_distances_t",
    "compute_displacements",
    "compute_center_of_mass",
    "compute_center_of_geometry",
    "find_closest_contact",
]


def compute_distances_core(
    positions,
    atom_pairs,
    unitcell_vectors=None,
    periodic=True,
    opt=True,
):
    """Compute the distances between pairs of atoms in each frame.

    Parameters
    ----------
    positions : np.ndarray of shape=(n_frames, n_atoms, 3), dtype=float
        The positions of all atoms for a given trajectory.

    atom_pairs : np.ndarray of shape=(num_pairs, 2), dtype=int
        Each row gives the indices of two atoms involved in the interaction.

    unitcell_vectors : None or np.ndarray of shape(n_frames, 3 x 3), default=None
        A numpy array that specifies the box vectors for all frames for a trajectory.

    periodic : bool, default=True
        If `periodic` is True and the trajectory contains unitcell
        information, we will compute distances under the minimum image
        convention.

    opt : bool, default=True
        Use an optimized native library to calculate distances. Our optimized
        SSE minimum image convention calculation implementation is over 1000x
        faster than the naive numpy implementation.

    Returns
    -------

    distances : np.ndarray, shape=(n_frames, num_pairs), dtype=float
        The distance, in each frame, between each pair of atoms.
    """

    xyz = ensure_type(
        positions,
        dtype=np.float32,
        ndim=3,
        name="traj.xyz",
        shape=(None, None, 3),
        warn_on_cast=False,
    )
    pairs = ensure_type(
        atom_pairs,
        dtype=np.int32,
        ndim=2,
        name="atom_pairs",
        shape=(None, 2),
        warn_on_cast=False,
    )
    if not np.all(np.logical_and(pairs < positions.shape[1], pairs >= 0)):
        raise ValueError(f"atom_pairs must be between 0 and {positions.shape[0]}")

    if len(pairs) == 0:
        return np.zeros((len(xyz), 0), dtype=np.float32)

    if periodic and (unitcell_vectors is not None):
        box = ensure_type(
            unitcell_vectors,
            dtype=np.float32,
            ndim=3,
            name="unitcell_vectors",
            shape=(len(xyz), 3, 3),
            warn_on_cast=False,
        )

        # convert to angles
        unitcell_angles = []
        for fr_unitcell_vectors in unitcell_vectors:
            _, _, _, alpha, beta, gamma = box_vectors_to_lengths_and_angles(
                fr_unitcell_vectors[0],
                fr_unitcell_vectors[1],
                fr_unitcell_vectors[2],
            )
            unitcell_angles.append(np.array([alpha, beta, gamma]))

        orthogonal = np.allclose(np.array(unitcell_angles), 90)

        if opt:
            out = np.empty((xyz.shape[0], pairs.shape[0]), dtype=np.float32)
            _geometry._dist_mic(
                xyz,
                pairs,
                box.transpose(0, 2, 1).copy(),
                out,
                orthogonal,
            )
            return out

        else:
            return _distance_mic(xyz, pairs, box.transpose(0, 2, 1), orthogonal)

    # either there are no unitcell vectors or they dont want to use them
    if opt:
        out = np.empty((xyz.shape[0], pairs.shape[0]), dtype=np.float32)
        _geometry._dist(xyz, pairs, out)
        return out
    else:
        return _distance(xyz, pairs)


def compute_distances(traj, atom_pairs, periodic=True, opt=True):
    """Compute the distances between pairs of atoms in each frame.

    Parameters
    ----------
    traj : Trajectory
        An mtraj trajectory.
    atom_pairs : np.ndarray, shape=(num_pairs, 2), dtype=int
        Each row gives the indices of two atoms involved in the interaction.
    periodic : bool, default=True
        If `periodic` is True and the trajectory contains unitcell
        information, we will compute distances under the minimum image
        convention.
    opt : bool, default=True
        Use an optimized native library to calculate distances. Our optimized
        SSE minimum image convention calculation implementation is over 1000x
        faster than the naive numpy implementation.

    Returns
    -------
    distances : np.ndarray, shape=(n_frames, num_pairs), dtype=float
        The distance, in each frame, between each pair of atoms.
    """

    return compute_distances_core(
        traj.xyz,
        atom_pairs,
        unitcell_vectors=traj.unitcell_vectors,
        periodic=periodic,
        opt=opt,
    )


def compute_distances_t(traj, atom_pairs, time_pairs, periodic=True, opt=True):
    """Compute the distances between pairs of atoms at pairs of times.

    Parameters
    ----------
    traj : Trajectory
        An mtraj trajectory.
    atom_pairs : np.ndarray, shape=(num_atom_pairs, 2), dtype=int
        Each row gives the indices of two atoms involved in the interaction.
    time_pairs : nd.array, shape=(num_times, 2), dtype=int
        Each row gives the indices of two frames.
    periodic : bool, default=True
        If `periodic` is True and the trajectory contains unitcell
        information, we will compute distances under the minimum image
        convention.
    opt : bool, default=True
        Use an optimized native library to calculate distances. Our optimized
        SSE minimum image convention calculation implementation is over 1000x
        faster than the naive numpy implementation.

    Returns
    -------
    distances : np.ndarray, shape=(num_times, num_atom_pairs), dtype=float
        The distance between each pair of atoms at t=t1 and t=t2.
    """
    xyz = ensure_type(
        traj.xyz,
        dtype=np.float32,
        ndim=3,
        name="traj.xyz",
        shape=(None, None, 3),
        warn_on_cast=False,
    )
    pairs = ensure_type(
        atom_pairs,
        dtype=np.int32,
        ndim=2,
        name="atom_pairs",
        shape=(None, 2),
        warn_on_cast=False,
    )
    times = ensure_type(
        time_pairs,
        dtype=np.int32,
        ndim=2,
        name="time_pairs",
        shape=(None, 2),
        warn_on_cast=False,
    )

    if not np.all(np.logical_and(pairs < traj.n_atoms, pairs >= 0)):
        raise ValueError("atom_pairs must be between 0 and %d" % traj.n_atoms)

    if not np.all(np.logical_and(times < traj.n_frames, times >= 0)):
        raise ValueError("time_pairs must be between 0 and %d" % traj.n_frames)

    if len(pairs) == 0:
        return np.zeros((len(xyz), 0), dtype=np.float32)

    if periodic and traj._have_unitcell:
        box = ensure_type(
            traj.unitcell_vectors,
            dtype=np.float32,
            ndim=3,
            name="unitcell_vectors",
            shape=(len(xyz), 3, 3),
            warn_on_cast=False,
        )
        orthogonal = np.allclose(traj.unitcell_angles, 90)
        if opt:
            out = np.empty((times.shape[0], pairs.shape[0]), dtype=np.float32)
            _geometry._dist_mic_t(
                xyz,
                pairs,
                times,
                box.transpose(0, 2, 1).copy(),
                out,
                orthogonal,
            )
            out = out.reshape((times.shape[0], pairs.shape[0]))
            return out
        else:
            return _distance_mic_t(
                xyz,
                pairs,
                times,
                box.transpose(0, 2, 1),
                orthogonal,
            )

    # either there are no unitcell vectors or they dont want to use them
    if opt:
        out = np.empty((times.shape[0], pairs.shape[0]), dtype=np.float32)
        _geometry._dist_t(xyz, pairs, times, out)
        return out
    else:
        return _distance_t(xyz, pairs, times)


def compute_displacements(traj, atom_pairs, periodic=True, opt=True):
    """Compute the displacement vector between pairs of atoms in each frame of a trajectory.

    Parameters
    ----------
    traj : Trajectory
        Trajectory to compute distances in
    atom_pairs : np.ndarray, shape[num_pairs, 2], dtype=int
        Each row gives the indices of two atoms.
    periodic : bool, default=True
        If `periodic` is True and the trajectory contains unitcell
        information, we will compute distances under the minimum image
        convention.
    opt : bool, default=True
        Use an optimized native library to calculate distances. Our
        optimized minimum image convention calculation implementation is
        over 1000x faster than the naive numpy implementation.

    Returns
    -------
    displacements : np.ndarray, shape=[n_frames, n_pairs, 3], dtype=float32
         The displacememt vector, in each frame, between each pair of atoms.
    """
    xyz = ensure_type(
        traj.xyz,
        dtype=np.float32,
        ndim=3,
        name="traj.xyz",
        shape=(None, None, 3),
        warn_on_cast=False,
    )
    pairs = ensure_type(
        np.asarray(atom_pairs),
        dtype=np.int32,
        ndim=2,
        name="atom_pairs",
        shape=(None, 2),
        warn_on_cast=False,
    )
    if not np.all(np.logical_and(pairs < traj.n_atoms, pairs >= 0)):
        raise ValueError("atom_pairs must be between 0 and %d" % traj.n_atoms)
    if len(pairs) == 0:  # If pairs is an empty slice of an array
        return np.zeros((len(xyz), 0, 3), dtype=np.float32)

    if periodic and traj._have_unitcell:
        box = ensure_type(
            traj.unitcell_vectors,
            dtype=np.float32,
            ndim=3,
            name="unitcell_vectors",
            shape=(len(xyz), 3, 3),
            warn_on_cast=False,
        )
        orthogonal = np.allclose(traj.unitcell_angles, 90)
        if opt:
            out = np.empty((xyz.shape[0], pairs.shape[0], 3), dtype=np.float32)
            _geometry._dist_mic_displacement(
                xyz,
                pairs,
                box.transpose(0, 2, 1).copy(),
                out,
                orthogonal,
            )
            return out
        else:
            return _displacement_mic(xyz, pairs, box.transpose(0, 2, 1), orthogonal)

    # either there are no unitcell vectors or they dont want to use them
    if opt:
        out = np.empty((xyz.shape[0], pairs.shape[0], 3), dtype=np.float32)
        _geometry._dist_displacement(xyz, pairs, out)
        return out
    return _displacement(xyz, pairs)


def compute_center_of_mass(traj, select=None):
    """Compute the center of mass for each frame.

    Parameters
    ----------
    traj : Trajectory
        Trajectory to compute center of mass for
    select : str, optional, default=all
        a mdtraj.Topology selection string that
        defines the set of atoms of which to calculate
        the center of mass, the default is all atoms

    Returns
    -------
    com : np.ndarray, shape=(n_frames, 3)
         Coordinates of the center of mass for each frame
    """
    com = np.empty((traj.n_frames, 3))

    if select is None:
        masses = np.array([a.element.mass for a in traj.top.atoms])
        masses /= masses.sum()

        xyz = traj.xyz

    else:
        atoms_of_interest = traj.topology.select(select)

        masses = np.array([traj.top.atom(i).element.mass for i in atoms_of_interest])
        masses /= masses.sum()

        xyz = traj.xyz[:, atoms_of_interest]

    for i, x in enumerate(xyz):
        com[i, :] = x.astype("float64").T.dot(masses)
    return com


def compute_center_of_geometry(traj):
    """Compute the center of geometry for each frame.

    Parameters
    ----------
    traj : Trajectory
        Trajectory to compute center of geometry for

    Returns
    -------
    com : np.ndarray, shape=(n_frames, 3)
         Coordinates of the center of geometry for each frame

    """

    centers = np.zeros((traj.n_frames, 3))

    for i, x in enumerate(traj.xyz):
        centers[i, :] = x.astype("float64").T.mean(axis=1)
    return centers


def find_closest_contact(traj, group1, group2, frame=0, periodic=True):
    """Find the closest contact between two groups of atoms.

    Given a frame of a Trajectory and two groups of atoms, identify the pair of
    atoms (one from each group) that form the closest contact between the two groups.

    Parameters
    ----------
    traj : Trajectory
        An mtraj trajectory.
    group1 : np.ndarray, shape=(num_atoms), dtype=int
        The indices of atoms in the first group.
    group2 : np.ndarray, shape=(num_atoms), dtype=int
        The indices of atoms in the second group.
    frame : int, default=0
        The frame of the Trajectory to take positions from
    periodic : bool, default=True
        If `periodic` is True and the trajectory contains unitcell
        information, we will compute distances under the minimum image
        convention.

    Returns
    -------
    result : tuple (int, int, float)
         The indices of the two atoms forming the closest contact, and the distance between them.
    """
    xyz = ensure_type(
        traj.xyz,
        dtype=np.float32,
        ndim=3,
        name="traj.xyz",
        shape=(None, None, 3),
        warn_on_cast=False,
    )[frame]
    atoms1 = ensure_type(
        group1,
        dtype=np.int32,
        ndim=1,
        name="group1",
        warn_on_cast=False,
    )
    atoms2 = ensure_type(
        group2,
        dtype=np.int32,
        ndim=1,
        name="group2",
        warn_on_cast=False,
    )
    if periodic and traj._have_unitcell:
        box = ensure_type(
            traj.unitcell_vectors,
            dtype=np.float32,
            ndim=3,
            name="unitcell_vectors",
            shape=(len(traj.xyz), 3, 3),
            warn_on_cast=False,
        )[frame]
    else:
        box = None
    return _geometry._find_closest_contact(xyz, atoms1, atoms2, box)


def _distance(xyz, pairs):
    "Distance between pairs of points in each frame"
    delta = np.diff(xyz[:, pairs], axis=2)[:, :, 0]
    return (delta**2.0).sum(-1) ** 0.5


def _distance_t(xyz, pairs, times):
    "Distance between pairs of points in specified frames"
    frame1 = xyz[:, pairs[:, 0]][times[:, 0]]
    frame2 = xyz[:, pairs[:, 1]][times[:, 1]]
    out = np.linalg.norm(frame1 - frame2, axis=2)

    return out


def _displacement(xyz, pairs):
    "Displacement vector between pairs of points in each frame"
    value = np.diff(xyz[:, pairs], axis=2)[:, :, 0]
    assert value.shape == (
        xyz.shape[0],
        pairs.shape[0],
        3,
    ), f"v.shape {str(value.shape)}, xyz.shape {str(xyz.shape)}, pairs.shape {str(pairs.shape)}"
    return value


def _reduce_box_vectors(vectors):
    """Make sure box vectors are in reduced form."""
    (bv1, bv2, bv3) = vectors
    bv3 -= bv2 * round(bv3[1] / bv2[1])
    bv3 -= bv1 * round(bv3[0] / bv1[0])
    bv2 -= bv1 * round(bv2[0] / bv1[0])
    return (bv1, bv2, bv3)


def _distance_mic(xyz, pairs, box_vectors, orthogonal):
    """Distance between pairs of points in each frame under the minimum image
    convention for periodic boundary conditions.

    The computation follows scheme B.9 in Tukerman, M. "Statistical
    Mechanics: Theory and Molecular Simulation", 2010.

    This is a slow pure python implementation, mostly for testing.
    """
    out = np.empty((xyz.shape[0], pairs.shape[0]), dtype=np.float32)
    for i in range(len(xyz)):
        bv1, bv2, bv3 = _reduce_box_vectors(box_vectors[i].T)

        for j, (a, b) in enumerate(pairs):
            r12 = xyz[i, b, :] - xyz[i, a, :]
            r12 -= bv3 * round(r12[2] / bv3[2])
            r12 -= bv2 * round(r12[1] / bv2[1])
            r12 -= bv1 * round(r12[0] / bv1[0])
            dist = np.linalg.norm(r12)
            if not orthogonal:
                for ii in range(-1, 2):
                    v1 = bv1 * ii
                    for jj in range(-1, 2):
                        v12 = bv2 * jj + v1
                        for kk in range(-1, 2):
                            new_r12 = r12 + v12 + bv3 * kk
                            dist = min(dist, np.linalg.norm(new_r12))
            out[i, j] = dist
    return out


def _distance_mic_t(xyz, pairs, times, box_vectors, orthogonal):
    """Distance between pairs of points between specified frames under the minimum image
    convention for periodic boundary conditions.

    The computation is modified from scheme B.9 in Tukerman, M. "Statistical
    Mechanics: Theory and Molecular Simulation", 2010.

    This is a slow pure python implementation, mostly for testing.
    """
    out = np.empty((times.shape[0], pairs.shape[0]), dtype=np.float32)
    for i, (a, b) in enumerate(times):
        bv1, bv2, bv3 = _reduce_box_vectors(box_vectors[a].T)
        for j, (c, d) in enumerate(pairs):
            r12 = xyz[a, c] - xyz[b, d]
            r12 -= bv3 * round(r12[2] / bv3[2])
            r12 -= bv2 * round(r12[1] / bv2[1])
            r12 -= bv1 * round(r12[0] / bv1[0])
            dist = np.linalg.norm(r12)
            if not orthogonal:
                for ii in range(-1, 2):
                    v1 = bv1 * ii
                    for jj in range(-1, 2):
                        v12 = bv2 * jj + v1
                        for kk in range(-1, 2):
                            new_r12 = r12 + v12 + bv3 * kk
                            dist = min(dist, np.linalg.norm(new_r12))
            out[i, j] = dist

    return out


def _displacement_mic(xyz, pairs, box_vectors, orthogonal):
    """Displacement vector between pairs of points in each frame under the
    minimum image convention for periodic boundary conditions.

    The computation follows scheme B.9 in Tukerman, M. "Statistical
    Mechanics: Theory and Molecular Simulation", 2010.

    This is a very slow pure python implementation, mostly for testing.
    """
    out = np.empty((xyz.shape[0], pairs.shape[0], 3), dtype=np.float32)
    for i in range(len(xyz)):
        bv1, bv2, bv3 = _reduce_box_vectors(box_vectors[i].T)
        # hinv, not used
        _ = np.linalg.inv(np.array([bv1, bv2, bv3]).T)

        for j, (a, b) in enumerate(pairs):
            r12 = xyz[i, b, :] - xyz[i, a, :]
            r12 -= bv3 * round(r12[2] / bv3[2])
            r12 -= bv2 * round(r12[1] / bv2[1])
            r12 -= bv1 * round(r12[0] / bv1[0])
            min_disp = r12
            dist2 = (r12 * r12).sum()
            if not orthogonal:
                for ii in range(-1, 2):
                    v1 = bv1 * ii
                    for jj in range(-1, 2):
                        v12 = bv2 * jj + v1
                        for kk in range(-1, 2):
                            tmp = r12 + v12 + bv3 * kk
                            new_dist2 = (tmp * tmp).sum()
                            if new_dist2 < dist2:
                                dist2 = new_dist2
                                min_disp = tmp
            out[i, j] = min_disp

    return out



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import os
import tempfile

import numpy as np
import pytest

import mdtraj as md
from mdtraj.core.topology import Topology
from mdtraj.formats import HDF5TrajectoryFile
from mdtraj.testing import eq

try:
    from openmm import unit as units

    HAVE_UNITS = True
except ImportError:
    HAVE_UNITS = False

needs_units = pytest.mark.skipif(not HAVE_UNITS, reason="requires openmm.units")


fd, temp = tempfile.mkstemp(suffix=".h5")
rng = np.random.default_rng()


def teardown_module(module):
    """remove the temporary file created by tests in this file
    this gets automatically called by pytest"""
    os.close(fd)
    os.unlink(temp)


def test_write_coordinates():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates)

    with HDF5TrajectoryFile(temp) as f:
        assert eq(f.root.coordinates[:], coordinates)
        assert eq(str(f.root.coordinates.attrs["units"]), "nanometers")


def test_write_coordinates_reshape():
    coordinates = rng.standard_normal((10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates)

    with HDF5TrajectoryFile(temp) as f:
        assert eq(f.root.coordinates[:], coordinates.reshape(1, 10, 3))
        assert eq(str(f.root.coordinates.attrs["units"]), "nanometers")


def test_write_multiple():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates)
        f.write(coordinates)

    with HDF5TrajectoryFile(temp) as f:
        assert eq(f.root.coordinates[:], np.vstack((coordinates, coordinates)))


def test_write_inconsistent():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates)
        # since the first frames we saved didn't contain velocities, we
        # can't save more velocities
        with pytest.raises(ValueError):
            f.write(coordinates, velocities=coordinates)


def test_write_inconsistent_2():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates, velocities=coordinates)
        # we're saving a deficient set of data, since before we wrote
        # more information.
        with pytest.raises(ValueError):
            f.write(coordinates)


@needs_units
def test_write_units():
    # openmm.units are automatically converted into MD units for storage on disk
    coordinates = units.Quantity(rng.standard_normal((4, 10, 3)), units.angstroms)
    velocities = units.Quantity(rng.standard_normal((4, 10, 3)), units.angstroms / units.year)

    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates, velocities=velocities)

    with HDF5TrajectoryFile(temp) as f:
        assert eq(f.root.coordinates[:], coordinates.value_in_unit(units.nanometers))
        assert eq(str(f.root.coordinates.attrs["units"]), "nanometers")

        assert eq(
            f.root.velocities[:],
            velocities.value_in_unit(units.nanometers / units.picosecond),
        )
        assert eq(str(f.root.velocities.attrs["units"]), "nanometers/picosecond")


def test_write_units2():
    from mdtraj.utils import unit

    coordinates = unit.quantity.Quantity(
        rng.standard_normal((4, 10, 3)),
        unit.unit_definitions.angstroms,
    )
    velocities = unit.quantity.Quantity(
        rng.standard_normal((4, 10, 3)),
        unit.unit_definitions.angstroms / unit.unit_definitions.year,
    )

    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates, velocities=velocities)

    with HDF5TrajectoryFile(temp) as f:
        assert eq(
            f.root.coordinates[:],
            coordinates.value_in_unit(unit.unit_definitions.nanometers),
        )
        assert eq(str(f.root.coordinates.attrs["units"]), "nanometers")

        assert eq(
            f.root.velocities[:],
            velocities.value_in_unit(
                unit.unit_definitions.nanometers / unit.unit_definitions.picosecond,
            ),
        )
        assert eq(str(f.root.velocities.attrs["units"]), "nanometers/picosecond")


@needs_units
def test_write_units_mismatch():
    velocities = units.Quantity(
        rng.standard_normal((4, 10, 3)),
        units.angstroms / units.picosecond,
    )

    with HDF5TrajectoryFile(temp, "w") as f:
        # if you try to write coordinates that are unitted and not
        # in the correct units, we find that
        with pytest.raises(TypeError):
            f.write(coordinates=velocities)


def test_topology(get_fn):
    top = md.load_pdb(get_fn("native.pdb")).topology

    with HDF5TrajectoryFile(temp, "w") as f:
        f.topology = top

    with HDF5TrajectoryFile(temp) as f:
        assert f.topology == top


def test_constraints():
    c = np.array(
        [(1, 2, 3.5)],
        dtype=np.dtype(
            [("atom1", np.int32), ("atom2", np.int32), ("distance", np.float32)],
        ),
    )

    with HDF5TrajectoryFile(temp, "w") as f:
        f.constraints = c

    with HDF5TrajectoryFile(temp) as f:
        assert eq(f.constraints, c)


def test_constraints2():
    c = np.array(
        [(1, 2, 3.5)],
        dtype=np.dtype(
            [("atom1", np.int32), ("atom2", np.int32), ("distance", np.float32)],
        ),
    )

    with HDF5TrajectoryFile(temp, "w") as f:
        f.constraints = c
        f.constraints = c

    with HDF5TrajectoryFile(temp) as f:
        assert eq(f.constraints, c)


def test_read_0():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates, alchemicalLambda=np.array([1, 2, 3, 4]))

    with HDF5TrajectoryFile(temp) as f:
        got = f.read()
        assert eq(got.coordinates, coordinates)
        assert eq(got.velocities, None)
        assert eq(got.alchemicalLambda, np.array([1, 2, 3, 4]))


@needs_units
def test_read_1():
    coordinates = units.Quantity(rng.standard_normal((4, 10, 3)), units.angstroms)
    velocities = units.Quantity(
        rng.standard_normal((4, 10, 3)),
        units.angstroms / units.years,
    )

    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates, velocities=velocities)

    with HDF5TrajectoryFile(temp) as f:
        got = f.read()
        assert eq(got.coordinates, coordinates.value_in_unit(units.nanometers))
        assert eq(
            got.velocities,
            velocities.value_in_unit(units.nanometers / units.picoseconds),
        )


def test_read_slice_0():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates, alchemicalLambda=np.array([1, 2, 3, 4]))

    with HDF5TrajectoryFile(temp) as f:
        got = f.read(n_frames=2)
        assert eq(got.coordinates, coordinates[:2])
        assert eq(got.velocities, None)
        assert eq(got.alchemicalLambda, np.array([1, 2]))


def test_read_slice_1():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates)

    with HDF5TrajectoryFile(temp) as f:
        got = f.read(n_frames=2)
        assert eq(got.coordinates, coordinates[:2])
        assert eq(got.velocities, None)

        got = f.read(n_frames=2)
        assert eq(got.coordinates, coordinates[2:])
        assert eq(got.velocities, None)


def test_read_slice_2():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates, alchemicalLambda=np.arange(4))

    with HDF5TrajectoryFile(temp) as f:
        got = f.read(atom_indices=np.array([0, 1]))
        assert eq(got.coordinates, coordinates[:, [0, 1], :])
        assert eq(got.alchemicalLambda, np.arange(4))


def test_read_slice_3():
    coordinates = rng.standard_normal((4, 10, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(coordinates, alchemicalLambda=np.arange(4))

    with HDF5TrajectoryFile(temp) as f:
        got = f.read(stride=2, atom_indices=np.array([0, 1]))
        assert eq(got.coordinates, coordinates[::2, [0, 1], :])
        assert eq(got.alchemicalLambda, np.arange(4)[::2])


def test_do_overwrite():
    with open(temp, "w") as f:
        f.write("a")

    with HDF5TrajectoryFile(temp, "w", force_overwrite=True) as f:
        f.write(rng.standard_normal((10, 5, 3)))


def test_vsite_elements(get_fn):
    #  Test case for issue #265
    pdb_filename = get_fn("GG-tip4pew.pdb")
    trj = md.load(pdb_filename)
    trj.save_hdf5(temp)

    md.load(temp, top=pdb_filename)


def test_dont_overwrite():
    with open(temp, "w") as f:
        f.write("a")
    with pytest.raises(IOError):
        with HDF5TrajectoryFile(temp, "w", force_overwrite=False) as f:
            f.write(rng.standard_normal((10, 5, 3)))


def test_attributes():
    constraints = np.zeros(
        10,
        dtype=[("atom1", np.int32), ("atom2", np.int32), ("distance", np.float32)],
    )
    with HDF5TrajectoryFile(temp, "w") as f:
        f.title = "mytitle"
        f.reference = "myreference"
        f.forcefield = "amber99"
        f.randomState = "sdf"
        f.application = "openmm"
        f.constraints = constraints

    with HDF5TrajectoryFile(temp) as g:
        eq(g.title, "mytitle")
        eq(g.reference, "myreference")
        eq(g.forcefield, "amber99")
        eq(g.randomState, "sdf")
        eq(g.application, "openmm")
        eq(g.constraints, constraints)


def test_append():
    x1 = rng.standard_normal((10, 5, 3))
    x2 = rng.standard_normal((8, 5, 3))
    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(x1)
    with HDF5TrajectoryFile(temp, "a") as f:
        f.write(x2)

    with HDF5TrajectoryFile(temp) as f:
        eq(f.root.coordinates[:], np.concatenate((x1, x2)))


def test_topology_None(h5traj):
    with HDF5TrajectoryFile(temp, "w") as f:
        f.topology = None

        assert f.topology is None, "The None argument did not pass through the topology setter properly"

    _, in_fn, _ = h5traj
    with HDF5TrajectoryFile(in_fn, "a") as f:
        # The file previously has a topology
        assert isinstance(f.topology, Topology), "The test HDF5 File does not contain a topology"
        f.topology = None

        # The topology should now be overwritten as None now
        assert f.topology is None, "The topology of the HDF5 file was not deleted"


def test_hdf5_bond_metadata(get_fn):
    # test that bond metadata is preserved when writing
    # and reading HDF5. Added in PR 2101

    traj = md.load(get_fn("imatinib.pdb"), bond_orders=True)

    with HDF5TrajectoryFile(temp, "w", bond_metadata=True) as f:
        f.write(traj.xyz)
        f.topology = traj.topology

    # Check that bond metadata was saved correctly
    with HDF5TrajectoryFile(temp) as f:
        top = f.topology
        for bond1, bond2 in zip(traj.topology.bonds, top.bonds):
            assert bond1 == bond2


def test_hdf5_formal_charge(get_fn):
    # test that atom formal charge is preserved when writing
    # and reading HDF5. Added in PR 2101

    traj = md.load(get_fn("1ply_charge.pdb"))

    with HDF5TrajectoryFile(temp, "w") as f:
        f.write(traj.xyz)
        f.topology = traj.topology

    # Check that formal charges were saved correctly
    with HDF5TrajectoryFile(temp) as f:
        top = f.topology
        for atom1, atom2 in zip(traj.topology.atoms, top.atoms):
            assert atom1.formal_charge == atom2.formal_charge


def test_hdf5_roundtrip(h5traj_full_metadata):
    # test that formal_charge and bond order/type is preserved
    # on topology when saving to and reading from HDF5.

    traj, _, tmp_dir = h5traj_full_metadata
    out_fn = os.path.join(tmp_dir, "roundtrip.h5")

    traj.save(out_fn)
    traj2 = md.load(out_fn)

    assert eq(traj.xyz, traj2.xyz)
    assert traj.topology == traj2.topology
    assert eq(traj.time, traj2.time)
    assert eq(traj.unitcell_lengths, traj2.unitcell_lengths)
    assert eq(traj.unitcell_angles, traj2.unitcell_angles)
    for atom1, atom2 in zip(traj.topology.atoms, traj2.topology.atoms):
        assert atom1.formal_charge == atom2.formal_charge
    for bond1, bond2 in zip(traj.topology.bonds, traj2.topology.bonds):
        assert bond1.order == bond2.order
        assert bond1.type == bond2.type



from mdtraj.formats.amberrst import AmberNetCDFRestartFile, AmberRestartFile
from mdtraj.formats.arc import ArcTrajectoryFile
from mdtraj.formats.dcd import DCDTrajectoryFile
from mdtraj.formats.dtr import DTRTrajectoryFile
from mdtraj.formats.gro import GroTrajectoryFile
from mdtraj.formats.hdf5 import HDF5TrajectoryFile
from mdtraj.formats.lammpstrj import LAMMPSTrajectoryFile
from mdtraj.formats.lh5 import LH5TrajectoryFile
from mdtraj.formats.mdcrd import MDCRDTrajectoryFile
from mdtraj.formats.netcdf import NetCDFTrajectoryFile
from mdtraj.formats.pdb import PDBTrajectoryFile
from mdtraj.formats.pdbx import PDBxTrajectoryFile
from mdtraj.formats.trr import TRRTrajectoryFile
from mdtraj.formats.xtc import XTCTrajectoryFile
from mdtraj.formats.xyzfile import XYZTrajectoryFile

__all__ = (
    "AmberNetCDFRestartFile",
    "AmberRestartFile",
    "ArcTrajectoryFile",
    "DCDTrajectoryFile",
    "DTRTrajectoryFile",
    "GroTrajectoryFile",
    "HDF5TrajectoryFile",
    "LAMMPSTrajectoryFile",
    "LH5TrajectoryFile",
    "MDCRDTrajectoryFile",
    "NetCDFTrajectoryFile",
    "PDBTrajectoryFile",
    "PDBxTrajectoryFile",
    "TRRTrajectoryFile",
    "XTCTrajectoryFile",
    "XYZTrajectoryFile",
)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


"""The mdtraj package contains tools for loading and saving molecular dynamics
trajectories in a variety of formats, including Gromacs XTC & TRR, CHARMM/NAMD
DCD, PDB, and HDF5.
"""

# silence cython related numpy warnings, see github.com/numpy/numpy/pull/432
import numpy as _  # noqa
from mdtraj.core import element
from mdtraj.core.topology import Amide, Aromatic, Double, Single, Topology, Triple
from mdtraj.core.trajectory import (
    Trajectory,
    iterload,
    join,
    load,
    load_frame,
    load_topology,
    open,
)
from mdtraj.formats.amberrst import load_ncrestrt, load_restrt
from mdtraj.formats.arc import load_arc
from mdtraj.formats.dcd import load_dcd
from mdtraj.formats.dtr import load_dtr, load_stk
from mdtraj.formats.hdf5 import load_hdf5
from mdtraj.formats.hoomdxml import load_hoomdxml
from mdtraj.formats.lammpstrj import load_lammpstrj
from mdtraj.formats.lh5 import load_lh5
from mdtraj.formats.mdcrd import load_mdcrd
from mdtraj.formats.mol2 import load_mol2
from mdtraj.formats.netcdf import load_netcdf
from mdtraj.formats.openmmxml import load_xml
from mdtraj.formats.pdb import load_pdb
from mdtraj.formats.prmtop import load_prmtop
from mdtraj.formats.psf import load_psf
from mdtraj.formats.registry import FormatRegistry
from mdtraj.formats.trr import load_trr
from mdtraj.formats.xtc import load_xtc
from mdtraj.formats.xyzfile import load_xyz
from mdtraj.geometry import (
    acylindricity,
    asphericity,
    baker_hubbard,
    compute_angles,
    compute_center_of_geometry,
    compute_center_of_mass,
    compute_chi1,
    compute_chi2,
    compute_chi3,
    compute_chi4,
    compute_chi5,
    compute_contacts,
    compute_dihedrals,
    compute_directors,
    compute_displacements,
    compute_distances,
    compute_distances_t,
    compute_drid,
    compute_dssp,
    compute_gyration_tensor,
    compute_inertia_tensor,
    compute_neighborlist,
    compute_neighbors,
    compute_nematic_order,
    compute_omega,
    compute_phi,
    compute_psi,
    compute_rdf,
    compute_rdf_t,
    compute_rg,
    density,
    dipole_moments,
    find_closest_contact,
    isothermal_compressability_kappa_T,
    kabsch_sander,
    pi_stacking,
    principal_moments,
    relative_shape_antisotropy,
    shrake_rupley,
    static_dielectric,
    thermal_expansion_alpha_P,
    wernet_nilsson,
)
from mdtraj.nmr import (
    chemical_shifts_ppm,
    chemical_shifts_shiftx2,
    chemical_shifts_spartaplus,
    compute_chemical_shifts,
    compute_J3_HN_C,
    compute_J3_HN_CB,
    compute_J3_HN_HA,
    reindex_dataframe_by_atoms,
)

from . import reporters
from ._lprmsd import lprmsd
from ._rmsd import rmsd, rmsf

from . import _version

__version__ = _version.get_versions()["version"]

__all__ = (
    "reporters",
    "lprmsd",
    "rmsd",
    "rmsf",
    "element",
    "Amide",
    "Aromatic",
    "Double",
    "Single",
    "Topology",
    "Triple",
    "load_ncrestrt",
    "load_restrt",
    "load_arc",
    "load_dcd",
    "load_dtr",
    "load_stk",
    "load_hdf5",
    "load_hoomdxml",
    "load_lammpstrj",
    "load_lh5",
    "load_mdcrd",
    "load_mol2",
    "load_netcdf",
    "load_xml",
    "load_pdb",
    "load_prmtop",
    "load_psf",
    "load_trr",
    "load_xtc",
    "load_xyz",
    "FormatRegistry",
    "open",
    "load",
    "iterload",
    "load_frame",
    "load_topology",
    "join",
    "Trajectory",
    "baker_hubbard",
    "shrake_rupley",
    "kabsch_sander",
    "compute_distances",
    "compute_distances_t",
    "compute_displacements",
    "compute_angles",
    "compute_dihedrals",
    "compute_phi",
    "compute_psi",
    "compute_chi1",
    "compute_chi2",
    "compute_chi3",
    "compute_chi4",
    "compute_chi5",
    "compute_omega",
    "compute_rg",
    "compute_contacts",
    "compute_drid",
    "compute_center_of_mass",
    "compute_center_of_geometry",
    "wernet_nilsson",
    "pi_stacking",
    "compute_dssp",
    "compute_neighbors",
    "compute_neighborlist",
    "compute_rdf",
    "compute_rdf_t",
    "compute_nematic_order",
    "compute_inertia_tensor",
    "compute_gyration_tensor",
    "find_closest_contact",
    "compute_directors",
    "principal_moments",
    "asphericity",
    "acylindricity",
    "relative_shape_antisotropy",
    "dipole_moments",
    "static_dielectric",
    "isothermal_compressability_kappa_T",
    "thermal_expansion_alpha_P",
    "density",
    "compute_J3_HN_C",
    "compute_J3_HN_CB",
    "compute_J3_HN_HA",
    "chemical_shifts_ppm",
    "chemical_shifts_shiftx2",
    "chemical_shifts_spartaplus",
    "compute_chemical_shifts",
    "reindex_dataframe_by_atoms",
)


def capi():
    import os
    import sys

    module_path = sys.modules["mdtraj"].__path__[0]
    return {
        "lib_dir": os.path.join(module_path, "core", "lib"),
        "include_dir": os.path.join(module_path, "core", "lib"),
    }



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Christian Schwantes
# Contributors: Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import itertools

import numpy as np
import pytest

import mdtraj as md
from mdtraj.testing import eq


def test_contact_0(get_fn):
    pdb = md.load(get_fn("bpti.pdb"))
    contacts = np.loadtxt(get_fn("contacts.dat")).astype(int)

    ca, ca_pairs = md.compute_contacts(pdb, contacts, scheme="ca")
    closest, closest_pairs = md.compute_contacts(pdb, contacts, scheme="closest")
    closest_heavy, closest_heavy_pairs = md.compute_contacts(
        pdb,
        contacts,
        scheme="closest-heavy",
    )
    sidechain, sidechain_pairs = md.compute_contacts(pdb, contacts, scheme="sidechain")
    with pytest.warns(UserWarning, match="selected topology includes at least one glycine"):
        sidechain_heavy, sidechain_heavy_pairs = md.compute_contacts(
            pdb,
            contacts,
            scheme="sidechain-heavy",
        )

    ref_ca = np.loadtxt(get_fn("cc_ca.dat"))
    ref_closest = np.loadtxt(get_fn("cc_closest.dat"))
    ref_closest_heavy = np.loadtxt(get_fn("cc_closest-heavy.dat"))
    ref_sidechain = np.loadtxt(get_fn("cc_sidechain.dat"))
    ref_sidechain_heavy = np.loadtxt(get_fn("cc_sidechain-heavy.dat"))

    eq(ref_ca, ca.flatten())
    eq(ref_closest, closest.flatten())
    eq(ref_closest_heavy, closest_heavy.flatten())
    eq(ref_sidechain, sidechain.flatten())
    eq(ref_sidechain_heavy, sidechain_heavy.flatten())
    eq(contacts, ca_pairs)
    eq(contacts, closest_pairs)
    eq(contacts, closest_heavy_pairs)
    eq(contacts, sidechain_pairs)
    eq(contacts, sidechain_heavy_pairs)


def test_contact_1(get_fn):
    pdb = md.load(get_fn("bpti.pdb"))
    dists, pairs = md.compute_contacts(pdb)
    for r0, r1 in pairs:
        # are these valid residue indices?
        pdb.topology.residue(r0)
        pdb.topology.residue(r1)

        assert not (abs(r0 - r1) < 3)

    maps = md.geometry.squareform(dists, pairs)
    for i, (r0, r1) in enumerate(pairs):
        for t in range(pdb.n_frames):
            eq(maps[t, r0, r1], dists[t, i])


def test_contact_2(get_fn):
    pdb = md.load(get_fn("1vii_sustiva_water.pdb"))
    dists, pairs = md.compute_contacts(pdb, scheme="closest")
    for r0, r1 in pairs:
        assert pdb.topology.residue(r0).name != "HOH"
        assert pdb.topology.residue(r1).name != "HOH"

    # spot check one of the pairs
    r0, r1 = pairs[10]
    atoms_r0 = [a.index for a in pdb.topology.residue(r0).atoms]
    atoms_r1 = [a.index for a in pdb.topology.residue(r1).atoms]

    atomdist = md.compute_distances(pdb, list(itertools.product(atoms_r0, atoms_r1)))

    np.testing.assert_array_equal(dists[:, 10], np.min(atomdist, axis=1))

    maps = md.geometry.squareform(dists, pairs)
    for i, (r0, r1) in enumerate(pairs):
        for t in range(pdb.n_frames):
            eq(maps[t, r0, r1], dists[t, i])


def test_contact_3(get_fn):
    pdb = md.load(get_fn("bpti.pdb"))
    beta = 20
    with np.errstate(divide="ignore", over="ignore"):
        dists, pairs = md.compute_contacts(pdb, soft_min=True, soft_min_beta=beta)
        maps = md.geometry.squareform(dists, pairs)

        for i, (r0, r1) in enumerate(pairs):
            for t in range(pdb.n_frames):
                assert np.allclose(
                    beta / np.log(np.sum(np.exp(beta / maps[t, r0, r1]))),
                    dists[t, i],
                )


def test_contact_4(get_fn):
    pdb = md.load(
        get_fn("1am7_protein.pdb"),
    )  # protonated and including at least one glycine residue (which has no heavy atoms in its sidechain)

    with pytest.warns(UserWarning, match="selected topology includes at least one glycine"):
        md.compute_contacts(
            pdb,
            contacts="all",
            scheme="sidechain-heavy",
        )  # test passes if this doesn't raise an exception



import re

import pytest

import mdtraj
from mdtraj.testing.docstrings import docstring_verifiers, import_all_modules

SKIP_MODULES = [
    r"mdtraj\.utils\.external",
    r"mdtraj\.utils\.six",
    r"mdtraj\.utils\.unit\.unit_math",
    r"mdtraj\.utils\.unit\.baseunit",
    r"mdtraj\.utils\.unit\.prefix",
    r"mdtraj\.utils\.unit\.unit",
    r"mdtraj\.utils\.unit\.quantity",
    r"mdtraj\.utils\.unit\.mymatrix",
    r"mdtraj\.formats\.lh5",
    r"mdtraj\.formats\.hdf5",
    r"mdtraj\.formats\.pdb\.pdbstructure",
    r"mdtraj\.scripts",
    r"mdtraj\.testing\.docscrape",
    r"mdtraj\.io",
    r"mdtraj\._version",
]

MODULES = import_all_modules(mdtraj)


@pytest.fixture(params=MODULES, ids=lambda x: x.__name__)
def module(request):
    mod = request.param
    if any(re.search(s, mod.__name__) for s in SKIP_MODULES):
        pytest.skip("Skip checking certain modules' docstrings")

    return mod


def test_docstrings(module):
    for test in docstring_verifiers(module):
        test()
